TEST_FILENAME = "test_netmhciipan_3_1_alleles.py"
ALLELES_FILENAME = "netmhciipan_3_1_alleles.txt"

special_chars = " *:-,/."
with open(TEST_FILENAME, "w") as f:
    f.write("from mhcgnomes import parse, Pair")
    with open(ALLELES_FILENAME) as alleles_file:
        for line in alleles_file:
            line = line.strip()
            if line.startswith("#") or not line:
                continue

            parts = line.split()
            allele_name = parts[0]
            fn_name = allele_name.replace('"', "").strip()
            for c in special_chars:
                fn_name = fn_name.replace(c, "_")
            fn_name = fn_name.replace("__", "_")

            f.write(f"\ndef test_{fn_name}():")

            f.write(f"\n    result = parse('{allele_name}', infer_class2_pairing=True)")
            f.write(f"""\n    assert result.__class__ is Pair, \\
                    'Expected parse(\"{allele_name}\") to be Pair but got %s' % (result,)""")
            f.write("\n")



import pytest

from mhcgnomes import Gene, Species, parse


@pytest.mark.parametrize(
    "gene_name",
    ["H", "J", "K", "L", "P", "V", "DRB6", "DRB7", "DRB8", "DRB9"],
)
def test_curated_human_pseudogene_loci(gene_name):
    gene = Gene.get("HLA", gene_name)

    assert gene.pseudogene_status is True
    assert gene.is_pseudogene


def test_human_gene_level_pseudogene_status_is_distinct_from_allele_annotation():
    allele = parse("HLA-H*02:01")

    assert allele.gene.pseudogene_status is True
    assert allele.gene.is_pseudogene
    assert not allele.annotation_pseudogene
    assert allele.is_pseudogene


def test_explicit_ps_annotation_still_marks_allele_without_gene_level_status():
    allele = parse("Caja-B5*01:01ps")

    assert allele.gene.pseudogene_status is None
    assert not allele.gene.is_pseudogene
    assert allele.annotation_pseudogene
    assert allele.is_pseudogene


def test_gene_pseudogene_status_is_species_specific_and_inherited():
    human_g = Gene.get("HLA", "G")
    rhesus_g = Gene.get("Mamu", "G")
    crab_eating_macaque_g = Gene.get("Mafa", "G")

    assert human_g.pseudogene_status is False
    assert not human_g.is_pseudogene
    assert rhesus_g.pseudogene_status is True
    assert rhesus_g.is_pseudogene
    assert crab_eating_macaque_g.pseudogene_status is True
    assert crab_eating_macaque_g.is_pseudogene


def test_descendant_specific_pseudogene_status_does_not_leak_to_sibling_species():
    bornean_orangutan_ap = Gene.get("Popy", "Ap")
    sumatran_orangutan_ap = Gene.get("Poab", "Ap")

    assert bornean_orangutan_ap.pseudogene_status is True
    assert bornean_orangutan_ap.is_pseudogene
    assert sumatran_orangutan_ap.pseudogene_status is None
    assert not sumatran_orangutan_ap.is_pseudogene


def test_caja_ps2_is_a_species_scoped_pseudogene():
    caja_ps2 = Gene.get("Caja", "PS2")
    generic_callithrix_ps2 = Gene.get("MaLA", "PS2")
    caja_ps2_allele = parse("Caja-PS2*01")

    assert caja_ps2.pseudogene_status is True
    assert caja_ps2.is_pseudogene
    assert caja_ps2_allele.gene == caja_ps2
    assert not caja_ps2_allele.annotation_pseudogene
    assert caja_ps2_allele.is_pseudogene
    assert generic_callithrix_ps2.pseudogene_status is None
    assert not generic_callithrix_ps2.is_pseudogene
    assert Gene.get("Capy", "PS2") is None


def test_gene_record_exposes_combined_pseudogene_status():
    hla_h_record = Gene.get("HLA", "H").to_record()
    hla_a_record = parse("HLA-A*02:01").to_record()
    annotated_record = parse("Caja-B5*01:01ps").to_record()

    assert hla_h_record["pseudogene_status"] is True
    assert hla_h_record["is_pseudogene"] is True
    assert hla_a_record["pseudogene_status"] is None
    assert hla_a_record["is_pseudogene"] is False
    assert annotated_record["pseudogene_status"] is True
    assert annotated_record["is_pseudogene"] is True


def test_gene_properties_and_families_are_immutable():
    human = Species.get("HLA")

    with pytest.raises(TypeError):
        human.gene_name_to_properties["H"] = {"pseudogene": False}
    with pytest.raises(TypeError):
        human.gene_name_to_properties["H"]["pseudogene"] = False
    with pytest.raises(TypeError):
        human.gene_family_name_to_gene_names["TAP"].add("TAPBP")


@pytest.mark.parametrize("species_prefix", ["HLA", "SLA", "Mamu", "Xetr"])
def test_tap_family_is_inherited_cross_species(species_prefix):
    species = Species.get(species_prefix)

    assert species.find_matching_gene_family_name("tap") == "TAP"
    assert species.get_gene_family_members("TAP") == ("TAP1", "TAP2")



from pathlib import Path
from time import time

import pandas as pd

from mhcgnomes import parse

DEFAULT_FASTA = Path(__file__).parents[1] / ".external-data" / "MHC_prot.fasta"


def run(n_repeats=3, filename=DEFAULT_FASTA):
    filename = str(filename)
    if filename.endswith("csv"):
        df = pd.read_csv(filename)
        alleles = list(df.allele.values)
    else:
        alleles = []
        with open(filename) as f:
            for l in f:
                if l.startswith(">"):
                    parts = l.split()
                    alleles.append(parts[1])
    n_alleles = len(alleles)
    print("Loaded %d alleles from %s" % (n_alleles, filename))
    for i in range(n_repeats):
        t0 = time()
        parsed = [parse(a) for a in alleles]
        t = time()
        elapsed = t - t0
        assert len(parsed) == len(alleles)
        print(
            "Iter #%d, parsed %d alleles in %0.2fs (%d alleles per second)"
            % (
                i + 1,
                n_alleles,
                elapsed,
                n_alleles / elapsed,
            )
        )


if __name__ == "__main__":
    run()



NETMHCPAN_3_0_DEST = "test_netmhcpan_3_0_alleles.py"
NETMHCPAN_3_0_SOURCE = "netmhcpan_3_0_alleles.txt"

NETMHCPAN_4_0_DEST = "test_netmhcpan_4_0_alleles.py"
NETMHCPAN_4_0_SOURCE = "netmhcpan_4_0_alleles.txt"

special_chars = " *:-,/."


def generate(src, dst, exclude=None):
    if exclude is None:
        exclude = set()
    alleles = set()
    with open(dst, "w") as f:
        f.write("from mhcgnomes import parse, Allele, Gene, AlleleWithoutGene\n")
        with open(src) as alleles_file:
            for line in alleles_file:
                line = line.strip()
                if line.startswith("#") or not line:
                    continue

                parts = line.split()
                allele_name = parts[0]
                if allele_name in exclude:
                    continue
                if allele_name in alleles:
                    print(f"Skipping repeat allele: '{allele_name}'")
                    continue
                alleles.add(allele_name)

                fn_name = allele_name.replace('"', "").strip()
                for c in special_chars:
                    fn_name = fn_name.replace(c, "_")
                fn_name = fn_name.replace("__", "_")

                f.write(f"\ndef test_{fn_name}():")

                f.write(f"\n    result = parse('{allele_name}')")
                if ":" in allele_name:
                    f.write(
                        f"""\n    assert result.__class__ is Allele, \\
                                'Expected parse(\"{allele_name}\") to be Allele but got %s' % (result,)"""
                    )
                else:
                    f.write(
                        f"""\n    assert result.__class__ in (Gene, Allele, AlleleWithoutGene), \\
                                'Unexpected type for parse(\"{allele_name}\"): %s' % (result,)"""
                    )
                f.write("\n")
    print(f"Wrote {len(alleles)} from {src} tests to {dst}")
    return alleles


netmhcpan_3_0_alleles = generate(src=NETMHCPAN_3_0_SOURCE, dst=NETMHCPAN_3_0_DEST)


generate(src=NETMHCPAN_4_0_SOURCE, dst=NETMHCPAN_4_0_DEST, exclude=netmhcpan_3_0_alleles)



"""
Parser support for Tasmanian devil (Sarcophilus harrisii) MHC class I names
from the DFT2 paper (Caldwell et al. 2018, eLife 7:e35314 / PMC6092122).

The paper uses a "SahaI*NN" shorthand where the number runs across multiple
class I loci (Saha-UA/UB/UC/UD). Table 1 assigns specific numbers to loci;
the rest are left unattributed. mhcgnomes routes paper names through a
curated alias table (see data/curated_allele_aliases.yaml) to either
locus-typed Alleles or AlleleWithoutGene values depending on paper
attribution.
"""

import pytest

from mhcgnomes import AlleleWithoutGene, AmbiguousAlleles, parse

from .common import eq_


@pytest.mark.parametrize(
    ("paper_name", "canonical"),
    [
        # Paper Table 1 explicit locus assignments
        ("SahaI*27", "Saha-UC*27"),
        ("SahaI*32", "Saha-UD*32"),
        ("SahaI*35", "Saha-UA*35"),
        ("SahaI*46", "Saha-UA*46"),
        ("SahaI*90", "Saha-UB*90"),
        # SahaI*27-1 is a single-substitution sibling of SahaI*27 at Saha-UC
        ("SahaI*27-1", "Saha-UC*27:01"),
        # The paper text uses en-dash (U+2013); folded to ASCII hyphen
        ("SahaI*27\u20131", "Saha-UC*27:01"),
        # Dashed species-gene form should resolve the same way
        ("Saha-I*27", "Saha-UC*27"),
        ("Saha-I*90", "Saha-UB*90"),
    ],
)
def test_assigned_sahai_names(paper_name, canonical):
    allele = parse(paper_name, use_allele_aliases=True)
    eq_(allele.to_string(), canonical)


@pytest.mark.parametrize(
    "paper_name",
    ["SahaI*29", "SahaI*33", "SahaI*36", "SahaI*37", "SahaI*97"],
)
def test_unassigned_sahai_names_become_allele_without_gene(paper_name):
    # Paper Table 1 does not attribute these to a specific Saha locus, so
    # the curated alias table routes them to an AlleleWithoutGene carrying
    # just the numeric designator.
    result = parse(paper_name, use_allele_aliases=True)
    assert isinstance(result, AlleleWithoutGene), (
        f"{paper_name!r} should produce AlleleWithoutGene, got {type(result).__name__}"
    )
    expected_number = paper_name.split("*")[1]
    eq_(result.name, expected_number)
    eq_(result.species.mhc_prefix, "Saha")


def test_saha_loci_parse_as_genes():
    # UA and UD were previously absent from species.yaml; they should now
    # parse as first-class Saha genes alongside UB/UC/UK/UM.
    for locus in ("UA", "UB", "UC", "UD", "UK", "UM"):
        gene = parse(f"Saha-{locus}")
        eq_(gene.to_string(), f"Saha-{locus}")


def test_saha_uc_sub_allele_suffix():
    # The numeric "-N" sub-allele suffix (paper uses this for SahaI*27-1)
    # should parse as a two-field allele regardless of whether the alias
    # table is involved.
    allele = parse("Saha-UC*27-1")
    eq_(allele.to_string(), "Saha-UC*27:01")


def test_em_dash_normalizes_to_hyphen():
    # Em-dash (U+2014) should fold to ASCII hyphen in the same way as
    # en-dash (U+2013), so round-tripping unicode-normalized paper text
    # produces the same parse.
    allele = parse("Saha-UC*27\u20141")
    eq_(allele.to_string(), "Saha-UC*27:01")


@pytest.mark.parametrize(
    ("paper_name", "expected_numbers"),
    [
        # Paper Table 1 slash-ambiguous designations: a single observed
        # DFT2 sequence that matches two IPD entries identical in the
        # region typed. Treated as two distinct alleles the input cannot
        # disambiguate between — not one allele with a slash in its name.
        ("SahaI*49/82", ("49", "82")),
        ("SahaI*74/88", ("74", "88")),
        ("Saha-I*74/88", ("74", "88")),
    ],
)
def test_slash_ambiguity_produces_two_alleles(paper_name, expected_numbers):
    result = parse(paper_name, use_allele_aliases=True)
    assert isinstance(result, AmbiguousAlleles), (
        f"{paper_name!r} should produce AmbiguousAlleles, got {type(result).__name__}"
    )
    eq_(result.num_alleles, 2)
    # alleles are sorted by value in ResultWithMultipleAlleles
    names = tuple(a.name for a in result.alleles)
    eq_(names, expected_numbers)
    for allele in result.alleles:
        assert isinstance(allele, AlleleWithoutGene)
        eq_(allele.species.mhc_prefix, "Saha")


def test_ambiguous_alleles_round_trip_string():
    result = parse("SahaI*74/88", use_allele_aliases=True)
    eq_(result.to_string(), "Saha-74/88")
    eq_(result.to_string(include_species=False), "74/88")


@pytest.mark.parametrize(
    "paper_name",
    ["SahaI*29", "SahaI*49", "SahaI*49/82", "SahaI*74/88"],
)
def test_class_i_context_preserved_without_locus(paper_name):
    # The "I" placeholder gene carries class I context; alias resolution
    # should thread that into AlleleWithoutGene / AmbiguousAlleles so
    # downstream class-based filters continue to work.
    result = parse(paper_name, use_allele_aliases=True)
    eq_(result.mhc_class, "I")
    assert result.is_class1
    if isinstance(result, AmbiguousAlleles):
        for allele in result.alleles:
            eq_(allele.mhc_class, "I")



#!/usr/bin/env python3
"""
Shared helpers for local release automation and GitHub Actions release checks.
"""

from __future__ import annotations

import argparse
import ast
import re
import sys
from collections.abc import Sequence
from pathlib import Path

VERSION_TAG_PREFIX = "v"
_VERSION_ALLOWED_RE = re.compile(r"^[0-9][0-9A-Za-z.+-]*$")


class ReleaseError(ValueError):
    pass


def _ensure(condition: bool, message: str) -> None:
    if not condition:
        raise ReleaseError(message)


def validate_version_string(version: str) -> str:
    value = version.strip()
    _ensure(bool(value), "Version must not be empty.")
    _ensure(
        bool(_VERSION_ALLOWED_RE.match(value)),
        f"Version '{value}' looks invalid (unexpected characters).",
    )
    return value


def parse_version_from_python_file(path: Path) -> str:
    if not path.exists():
        raise ReleaseError(f"Missing version file: {path}")
    if not path.is_file():
        raise ReleaseError(f"Version path is not a regular file: {path}")

    try:
        text = path.read_text(encoding="utf-8")
    except OSError as exc:
        raise ReleaseError(f"Failed to read version file {path}: {exc}") from exc

    try:
        tree = ast.parse(text, filename=str(path))
    except SyntaxError as exc:
        raise ReleaseError(f"Version file is not valid Python syntax: {path}: {exc}") from exc

    found: list[str] = []
    for node in tree.body:
        if isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == "__version__":
                    found.append(_extract_version_literal(node.value, path))
        elif (
            isinstance(node, ast.AnnAssign)
            and isinstance(node.target, ast.Name)
            and node.target.id == "__version__"
        ):
            if node.value is None:
                raise ReleaseError(f"{path}: __version__ is annotated but not assigned.")
            found.append(_extract_version_literal(node.value, path))

    _ensure(bool(found), f"Failed to find __version__ assignment in {path}.")
    _ensure(
        len(found) == 1,
        f"Multiple __version__ assignments found in {path}. Refuse to guess.",
    )
    return validate_version_string(found[0])


def _extract_version_literal(node: ast.expr, path: Path) -> str:
    if isinstance(node, ast.Constant) and isinstance(node.value, str):
        return node.value
    raise ReleaseError(f"{path}: __version__ must be a string literal.")


def tag_for_version(version: str, prefix: str = VERSION_TAG_PREFIX) -> str:
    clean_version = validate_version_string(version)
    _ensure(bool(prefix), "Tag prefix must not be empty.")
    return f"{prefix}{clean_version}"


def ensure_tag_matches_version(
    tag: str,
    version: str,
    *,
    prefix: str = VERSION_TAG_PREFIX,
) -> None:
    clean_tag = tag.strip()
    _ensure(bool(clean_tag), "Tag must not be empty.")
    expected = tag_for_version(version, prefix=prefix)
    _ensure(
        clean_tag == expected,
        f"Tag/version mismatch: expected '{expected}' for version '{version}', got '{clean_tag}'.",
    )


def check_tag_against_version_file(
    tag: str,
    version_file: Path,
    *,
    prefix: str = VERSION_TAG_PREFIX,
) -> str:
    version = parse_version_from_python_file(version_file)
    ensure_tag_matches_version(tag, version, prefix=prefix)
    return version


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="python -m release_utils")
    subparsers = parser.add_subparsers(dest="command", required=True)

    print_version = subparsers.add_parser("print-version")
    print_version.add_argument("--version-file", default="mhcgnomes/version.py")
    print_version.add_argument("--as-tag", action="store_true")

    check_tag = subparsers.add_parser("check-tag")
    check_tag.add_argument("--version-file", default="mhcgnomes/version.py")
    check_tag.add_argument("--tag", required=True)

    return parser


def main(argv: Sequence[str] | None = None) -> int:
    parser = build_parser()
    args = parser.parse_args(list(argv) if argv is not None else None)

    try:
        version_file = Path(args.version_file)
        version = parse_version_from_python_file(version_file)

        if args.command == "print-version":
            output = tag_for_version(version) if args.as_tag else version
            print(output)
            return 0

        if args.command == "check-tag":
            check_tag_against_version_file(args.tag, version_file)
            print(f"OK: tag '{args.tag}' matches version '{version}' from {version_file}")
            return 0
    except ReleaseError as exc:
        print(f"ERROR: {exc}", file=sys.stderr)
        return 1

    parser.error(f"Unhandled command: {args.command}")
    return 2


if __name__ == "__main__":
    raise SystemExit(main())



#!/usr/bin/env python3
"""
Collect MHC allele strings from all candidate papers.

For each paper in candidate_papers.tsv:
1. Look up PMC ID (if available) via Entrez
2. Download supplementary files from PMC OA or publisher
3. Scrape for MHC allele strings
4. Output consolidated validation TSV

Usage:
    python paper/scripts/collect_all.py
"""

import csv
import os
import re
import subprocess
import sys
import time
import urllib.request
import urllib.error
from pathlib import Path

try:
    from Bio import Entrez
except ImportError:
    print("Requires biopython: pip install biopython", file=sys.stderr)
    sys.exit(1)

Entrez.email = "mhcgnomes-paper@example.com"

ROOT = Path(__file__).resolve().parent.parent.parent
CANDIDATES = ROOT / "paper" / "validation" / "candidate_papers.tsv"
RAW_DIR = ROOT / "paper" / "raw"
VAL_DIR = ROOT / "paper" / "validation"
SCRAPE_SCRIPT = ROOT / "paper" / "scripts" / "scrape_paper.py"


def pmid_to_pmc(pmid):
    """Convert PMID to PMCID via Entrez."""
    try:
        handle = Entrez.elink(dbfrom="pubmed", db="pmc", id=pmid)
        result = Entrez.read(handle)
        handle.close()
        for linkset in result:
            for db_link in linkset.get("LinkSetDb", []):
                if db_link.get("DbTo") == "pmc":
                    links = db_link.get("Link", [])
                    if links:
                        return links[0]["Id"]
    except Exception:
        pass
    return None


def get_pmc_supp_urls(pmc_id):
    """Get supplementary file URLs from PMC OA service."""
    urls = []
    try:
        oa_url = (
            f"https://www.ncbi.nlm.nih.gov/pmc/utils/oa/oa.fcgi?"
            f"id=PMC{pmc_id}"
        )
        req = urllib.request.Request(oa_url, headers={"User-Agent": "mhcgnomes-paper/1.0"})
        with urllib.request.urlopen(req, timeout=15) as resp:
            text = resp.read().decode()
            # Parse XML for links
            for m in re.finditer(r'href="(https?://[^"]+)"', text):
                href = m.group(1)
                if any(ext in href.lower() for ext in [".xlsx", ".xls", ".csv", ".tsv", ".zip", ".tar"]):
                    urls.append(href)
    except Exception:
        pass

    # Also try the PMC supplementary page directly
    try:
        supp_url = f"https://www.ncbi.nlm.nih.gov/pmc/articles/PMC{pmc_id}/bin/"
        req = urllib.request.Request(supp_url, headers={"User-Agent": "mhcgnomes-paper/1.0"})
        with urllib.request.urlopen(req, timeout=15) as resp:
            text = resp.read().decode()
            for m in re.finditer(r'href="([^"]*\.(xlsx|xls|csv|tsv|txt|zip|docx))"', text, re.IGNORECASE):
                href = m.group(1)
                if not href.startswith("http"):
                    href = f"https://www.ncbi.nlm.nih.gov{href}" if href.startswith("/") else f"{supp_url}{href}"
                urls.append(href)
    except Exception:
        pass

    return list(set(urls))


def get_doi_supp_urls(doi):
    """Try to find supplementary files via DOI redirect."""
    if not doi:
        return []
    urls = []
    try:
        # Follow DOI redirect to get publisher page
        req = urllib.request.Request(
            f"https://doi.org/{doi}",
            headers={"User-Agent": "mhcgnomes-paper/1.0"}
        )
        with urllib.request.urlopen(req, timeout=15) as resp:
            final_url = resp.url
            text = resp.read().decode(errors="replace")

        # Look for supplementary file links
        for m in re.finditer(
            r'href="([^"]*(?:supplementa|supp|S\d+|table_S|Table_S|additional)[^"]*'
            r'\.(?:xlsx|xls|csv|tsv|zip|txt))"',
            text,
            re.IGNORECASE,
        ):
            href = m.group(1)
            if not href.startswith("http"):
                from urllib.parse import urljoin
                href = urljoin(final_url, href)
            urls.append(href)
    except Exception:
        pass
    return list(set(urls))


def download_file(url, dest):
    """Download a URL to a local path."""
    try:
        req = urllib.request.Request(url, headers={"User-Agent": "mhcgnomes-paper/1.0"})
        with urllib.request.urlopen(req, timeout=30) as resp:
            with open(dest, "wb") as f:
                f.write(resp.read())
        return True
    except Exception as e:
        print(f"    Download failed: {e}", file=sys.stderr)
        return False


def scrape_file(filepath, species, source, output):
    """Run scrape_paper.py on a file."""
    try:
        result = subprocess.run(
            [
                sys.executable, str(SCRAPE_SCRIPT),
                "--input", str(filepath),
                "--species", species,
                "--source", source,
                "--output", str(output),
            ],
            capture_output=True,
            text=True,
            timeout=60,
        )
        if result.returncode == 0:
            # Count lines
            lines = sum(1 for _ in open(output)) - 1  # subtract header
            return lines
        else:
            print(f"    Scrape error: {result.stderr[:200]}", file=sys.stderr)
    except Exception as e:
        print(f"    Scrape error: {e}", file=sys.stderr)
    return 0


def load_candidates():
    """Load candidate papers from TSV."""
    papers = []
    with open(CANDIDATES) as f:
        reader = csv.DictReader(f, delimiter="\t")
        for row in reader:
            papers.append(row)
    return papers


def main():
    papers = load_candidates()
    print(f"Processing {len(papers)} candidate papers...\n", file=sys.stderr)

    total_strings = 0
    papers_with_data = 0
    papers_no_supp = 0
    papers_no_mhc = 0

    all_validation_rows = []

    for i, paper in enumerate(papers):
        pmid = paper["pmid"]
        stratum = paper["stratum"]
        title = paper["title"][:70]
        doi = paper.get("doi", "")

        print(f"[{i+1}/{len(papers)}] PMID:{pmid} ({stratum}) {title}...", file=sys.stderr)

        # Create output directory
        safe_id = f"PMID_{pmid}"
        raw_dir = RAW_DIR / safe_id
        raw_dir.mkdir(parents=True, exist_ok=True)

        # Find supplementary files
        pmc_id = pmid_to_pmc(pmid)
        supp_urls = []
        if pmc_id:
            print(f"  PMC{pmc_id}", file=sys.stderr)
            supp_urls = get_pmc_supp_urls(pmc_id)

        if not supp_urls and doi:
            supp_urls = get_doi_supp_urls(doi)

        if not supp_urls:
            print(f"  No supplementary files found", file=sys.stderr)
            papers_no_supp += 1
            time.sleep(0.3)
            continue

        print(f"  Found {len(supp_urls)} supplementary URLs", file=sys.stderr)

        # Download and scrape each file
        paper_strings = 0
        for url in supp_urls[:5]:  # limit to 5 files per paper
            filename = url.split("/")[-1].split("?")[0]
            if not filename:
                filename = "supplement"
            dest = raw_dir / filename

            if not dest.exists():
                print(f"  Downloading {filename}...", file=sys.stderr)
                if not download_file(url, dest):
                    continue
            else:
                print(f"  Already have {filename}", file=sys.stderr)

            # Skip non-scrapeable files
            if dest.suffix.lower() in (".zip", ".tar", ".gz", ".pdf", ".docx", ".doc"):
                print(f"  Skipping {dest.suffix} file", file=sys.stderr)
                continue

            # Scrape
            out_tsv = VAL_DIR / f"{safe_id}_{dest.stem}.tsv"
            n = scrape_file(dest, "", f"PMID:{pmid}", out_tsv)
            if n > 0:
                print(f"  Scraped {n} MHC strings from {filename}", file=sys.stderr)
                paper_strings += n
                # Load and accumulate
                with open(out_tsv) as f:
                    reader = csv.DictReader(f, delimiter="\t")
                    for row in reader:
                        all_validation_rows.append(row)

        if paper_strings > 0:
            papers_with_data += 1
            total_strings += paper_strings
        else:
            papers_no_mhc += 1

        time.sleep(0.5)  # be polite

    # Write consolidated validation file
    consolidated = VAL_DIR / "all_papers.tsv"
    with open(consolidated, "w") as f:
        writer = csv.DictWriter(
            f,
            fieldnames=["raw_string", "expected_species", "source"],
            delimiter="\t",
        )
        writer.writeheader()
        writer.writerows(all_validation_rows)

    print(f"\n{'='*60}", file=sys.stderr)
    print(f"Results:", file=sys.stderr)
    print(f"  Papers processed:    {len(papers)}", file=sys.stderr)
    print(f"  With MHC data:       {papers_with_data}", file=sys.stderr)
    print(f"  No supplementary:    {papers_no_supp}", file=sys.stderr)
    print(f"  No MHC strings:      {papers_no_mhc}", file=sys.stderr)
    print(f"  Total MHC strings:   {total_strings}", file=sys.stderr)
    print(f"  Consolidated file:   {consolidated}", file=sys.stderr)


if __name__ == "__main__":
    main()



"""
Coverage of the two authorities that assign MHC names.

Species designations come from the Comparative MHC Nomenclature Committee via
IPD-MHC group listings; human gene names come from IMGT/HLA. These tests pin
the entries added from a sweep of both against the ontology.

https://github.com/pirl-unc/mhcgnomes/issues/111
https://github.com/pirl-unc/mhcgnomes/issues/113
"""

import pytest

from mhcgnomes import Species, parse

from .common import eq_

# ---------------------------------------------------------------------------
# IPD-MHC species designations
# ---------------------------------------------------------------------------

# (IPD prefix, latin name) — https://www.ebi.ac.uk/ipd/mhc/group/NHP/species
CAPUCHINS = [
    ("Ceal", "Cebus albifrons"),
    ("Ceap", "Cebus apella"),
    ("Ceca", "Cebus capucinus"),
    ("Ceim", "Cebus imitator"),
    ("Ceka", "Cebus kaapori"),
    ("Ceol", "Cebus olivaceus"),
    ("Saca", "Sapajus cay"),
    ("Sali", "Sapajus libidinosus"),
    ("Sama", "Sapajus apella macrocephalus"),
    ("Sani", "Sapajus nigritus"),
    ("Saxa", "Sapajus xanthosternos"),
]

# https://www.ebi.ac.uk/ipd/mhc/group/CeLA/species
CETACEANS = [
    ("Bamubr", "Balaenoptera musculus brevicauda"),
    ("Bari", "Balaenoptera ricei"),
    ("Esro", "Eschrichtius robustus"),
    ("Eugl", "Eubalaena glacialis"),
    ("Inge", "Inia geoffrensis"),
    ("Kosi", "Kogia sima"),
    ("Laal", "Lagenorhynchus albirostris"),
    ("Live", "Lipotes vexillifer"),
    ("Momo", "Monodon monoceros"),
    ("Phph", "Phocoena phocoena"),
    ("Phsi", "Phocoena sinus"),
]

# https://www.ebi.ac.uk/ipd/mhc/group/DLA/species
CANIDS = [
    ("Caad", "Canis adustus"),
    ("Caba", "Canis lupus baileyi"),
    ("Cala", "Canis latrans"),
    ("Calu", "Canis lupus"),
    ("Came", "Canis mesomelas"),
    ("Caru", "Canis rufus"),
    ("Casi", "Canis simensis"),
    ("Cual", "Cuon alpinus"),
    ("Lypi", "Lycaon pictus"),
]

# https://www.ebi.ac.uk/ipd/mhc/group/SLA/species
SUIDS = [
    ("Phaf", "Phacochoerus africanus"),
    ("Sudo", "Sus domesticus"),
    ("Susc", "Sus scrofa"),
]

IPD_SPECIES = CAPUCHINS + CETACEANS + CANIDS + SUIDS


@pytest.mark.parametrize("prefix,latin_name", IPD_SPECIES)
def test_ipd_prefix_resolves_to_its_species(prefix, latin_name):
    species = Species.get(prefix)
    assert species is not None, f"Species.get({prefix!r}) returned None"
    eq_(species.name, latin_name)


@pytest.mark.parametrize("prefix,latin_name", IPD_SPECIES)
def test_ipd_species_reachable_by_latin_name(prefix, latin_name):
    eq_(Species.get_by_latin_name(latin_name).prefix, prefix)


# Two IPD designations collide with a code already in published use by another
# species, and IPD holds no allele sequences for either of them. Their species
# are present and their designations recorded, but reserved rather than global.
RESERVED_DESIGNATIONS = [
    ("Caau", "Canis aureus", "Carassius auratus"),
    ("Hyam", "Hyperoodon ampullatus", "Hybognathus amarus"),
]


@pytest.mark.parametrize("prefix,ipd_species,attested_species", RESERVED_DESIGNATIONS)
def test_reserved_designation_species_is_present(prefix, ipd_species, attested_species):
    assert Species.get_by_latin_name(ipd_species) is not None


@pytest.mark.parametrize("prefix,ipd_species,attested_species", RESERVED_DESIGNATIONS)
def test_reserved_designation_does_not_take_the_prefix(prefix, ipd_species, attested_species):
    eq_(Species.get(prefix).name, attested_species)


# ---------------------------------------------------------------------------
# Genus boundaries
#
# Cuon, Lycaon and Phacochoerus are canids and suids but not Canis or Sus, so
# they must not inherit the DLA/SLA prefix or those genera's gene lists. That
# is the failure described in issue #109 for water buffalo under Bos sp.
# ---------------------------------------------------------------------------


@pytest.mark.parametrize(
    "latin_name,family_node",
    [
        ("Cuon alpinus", "Canidae sp."),
        ("Lycaon pictus", "Canidae sp."),
        ("Phacochoerus africanus", "Suidae sp."),
    ],
)
def test_out_of_genus_species_hang_off_the_family_not_the_genus(latin_name, family_node):
    eq_(Species.get_by_latin_name(latin_name).parent_species.name, family_node)


@pytest.mark.parametrize("prefix", ["Cual", "Lypi"])
def test_non_canis_canids_do_not_answer_to_DLA(prefix):
    species = Species.get(prefix)
    assert "DLA" not in {p.upper() for p in species.all_mhc_prefixes}


def test_warthog_does_not_answer_to_SLA():
    species = Species.get("Phaf")
    assert "SLA" not in {p.upper() for p in species.all_mhc_prefixes}


@pytest.mark.parametrize("prefix,owner", [("DLA", "Canis sp."), ("SLA", "Sus sp.")])
def test_family_node_does_not_take_over_the_group_prefix(prefix, owner):
    """A taxonomic prefix is not inherited, so DLA and SLA keep their owners."""
    eq_(Species.get(prefix).name, owner)


def test_existing_group_alleles_are_unaffected():
    eq_(parse("DLA-DRB1*001:01").species.name, "Canis sp.")
    eq_(parse("SLA-1*01:01").species.name, "Sus sp.")


# ---------------------------------------------------------------------------
# IMGT/HLA gene names
# ---------------------------------------------------------------------------

IMGT_HLA_GENES_ADDED = [
    "DRB2",
    "DQB3",
    "DPA2",
    "DPA3",
    "DPB2",
    "MICC",
    "MICD",
    "MICE",
    "PSMB8",
    "PSMB9",
]


@pytest.mark.parametrize("gene_name", IMGT_HLA_GENES_ADDED)
def test_imgt_hla_gene_parses(gene_name):
    result = parse(f"HLA-{gene_name}", raise_on_error=False)
    assert result is not None, f"HLA-{gene_name} did not parse"
    eq_(result.species.name, "Homo sapiens")
    eq_(result.name, gene_name)


@pytest.mark.parametrize(
    "gene_name",
    ["DRB2", "DQB3", "DPA2", "DPA3", "DPB2", "MICC", "MICD", "MICE"],
)
def test_added_imgt_pseudogenes_are_marked_as_such(gene_name):
    assert parse(f"HLA-{gene_name}").is_pseudogene


def test_proteasome_subunits_keep_their_legacy_names():
    """LMP7/LMP2 are to PSMB8/PSMB9 what RING4/RING11 are to TAP1/TAP2."""
    eq_(parse("HLA-LMP7").name, "PSMB8")
    eq_(parse("HLA-LMP2").name, "PSMB9")


def test_single_letter_haplotype_shorthand_is_not_shadowed():
    """
    IMGT/HLA also names the class I fragments N, R, S, T, U, W, X, Y, Z. They
    are deliberately not in the ontology: adding them turns bare mouse and rat
    haplotype shorthand into human gene fragments. See issue #113.
    """
    eq_(parse("n").to_string(), "RT1-n")
    eq_(parse("u").to_string(), "H2-u")
    eq_(parse("s").to_string(), "H2-s")
    assert parse("HLA-Z*01:01", raise_on_error=False) is None



import pandas as pd

TEST_FILENAME = "test_iedb_names.py"
df = pd.read_csv("iedb_allele_counts.csv")

alleles = set()
special_chars = " *:-,/."
with open(TEST_FILENAME, "w") as f:
    f.write("from mhcgnomes import parse\n")
    for allele_name, count in sorted(zip(df.allele, df["number_of_entries"])):
        if allele_name in alleles:
            print(f"Skipping repeat allele: '{allele_name}'")
            continue
        alleles.add(allele_name)
        fn_name = allele_name.replace('"', "").strip()
        for c in special_chars:
            fn_name = fn_name.replace(c, "_")
        fn_name = fn_name.replace("__", "_")

        f.write(f"\ndef test_{fn_name}():")
        f.write(f"\n    # {allele_name} occurs {count} times in 2019 IEDB snapshot")
        f.write(f'\n    assert parse("{allele_name}") is not None')
        f.write("\n")

print("Wrote %d alleles to %s" % (len(alleles), TEST_FILENAME))



"""
Tests for parsing fixes in v3.30.0.

Covers:
- H2-AA / H2-Aa parsing as Gene (not Pair)
- H2-AB1 / H2-Ab1 parsing as Gene (not parse error)
- RT1-B parsing as Class2Locus (not Haplotype)
- RT1-B gene alias (B → Bb) via preferred_result_types
- Nibea species recognition
- Regression: IAk-style haplotype+locus parsing still works
"""

from mhcgnomes import Class2Locus, Gene, Pair, Species, parse

from .common import eq_

# ---------------------------------------------------------------------------
# H2-AA / H2-Aa: should parse as Gene, not Pair
# ---------------------------------------------------------------------------


def test_h2_aa_parses_as_gene():
    result = parse("H2-AA")
    eq_(type(result), Gene)
    eq_(result.name, "AA")


def test_h2_aa_lowercase_parses_as_gene():
    result = parse("H2-Aa")
    eq_(type(result), Gene)
    eq_(result.name, "AA")


def test_h2_aa_species_is_mouse():
    result = parse("H2-AA")
    eq_(result.species, Species.get("H2"))


# ---------------------------------------------------------------------------
# H2-AB1 / H2-Ab1: should parse as Gene (was hard parse error)
# ---------------------------------------------------------------------------


def test_h2_ab1_parses_as_gene():
    result = parse("H2-AB1")
    eq_(type(result), Gene)
    eq_(result.name, "AB1")


def test_h2_ab1_lowercase_parses_as_gene():
    result = parse("H2-Ab1")
    eq_(type(result), Gene)
    eq_(result.name, "AB1")


# ---------------------------------------------------------------------------
# RT1-B: should parse as Class2Locus (not Haplotype)
# ---------------------------------------------------------------------------


def test_rt1_b_parses_as_class2_locus():
    result = parse("RT1-B")
    eq_(type(result), Class2Locus)
    eq_(result.name, "B")


def test_rt1_b_species_is_rat():
    result = parse("RT1-B")
    eq_(result.species, Species.get("RT1"))


def test_rt1_b_preferred_gene_returns_bb():
    """With preferred_result_types=[Gene], RT1-B should resolve to Gene Bb
    via the gene alias (NCBI GeneID 24738 legacy entry)."""
    result = parse("RT1-B", preferred_result_types=[Gene])
    eq_(type(result), Gene)
    eq_(result.name, "Bb")


# ---------------------------------------------------------------------------
# Nibea species recognition
# ---------------------------------------------------------------------------


def test_nibea_mitsukurii_species():
    species = Species.get("NibeMits")
    assert species is not None
    eq_(species.latin_name, "Nibea mitsukurii")


def test_nibea_albiflora_species():
    species = Species.get("NibeAlbi")
    assert species is not None
    eq_(species.latin_name, "Nibea albiflora")


def test_nibea_mitsukurii_full_name():
    species = Species.get("Nibea mitsukurii")
    assert species is not None
    eq_(species.prefix, "NibeMits")


# ---------------------------------------------------------------------------
# Regression: IAk-style haplotype+locus splitting still works
# ---------------------------------------------------------------------------


def test_h2_iak_still_parses_as_pair():
    """Ensure the guard in parse_haplotype_with_class2_locus_from_any_string_split
    doesn't break legitimate locus+haplotype splits like IAk.
    IAk = haplotype 'k' restricted to locus A, which collapses to a Pair."""
    result = parse("H2-IAk")
    eq_(type(result), Pair)
    eq_(result.alpha.gene.name, "AA")
    eq_(result.beta.gene.name, "AB")



from mhcgnomes import Allele, parse

from .common import eq_


def test_allele_get_A0201():
    allele = Allele.get("HLA", "A", "02", "01")
    assert allele is not None
    assert type(allele) is Allele
    eq_(allele.species_prefix, "HLA")
    eq_(allele.gene_name, "A")
    eq_(list(allele.allele_fields), ["02", "01"])
    eq_(allele.mhc_class, "Ia")


def test_restrict_num_allele_fields_A02010101():
    allele_eight_digit = Allele.get("HLA", "A", "02", "01", "01", "01")
    assert allele_eight_digit is not None
    assert type(allele_eight_digit) is Allele
    eq_(allele_eight_digit.num_allele_fields, 4)
    allele_four_digit = allele_eight_digit.restrict_allele_fields(2)
    eq_(allele_four_digit.num_allele_fields, 2)


def test_no_annotations():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01")
    assert not allele.annotation_null
    assert not allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert not allele.annotation_secreted
    assert not allele.annotation_pseudogene
    assert not allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert not allele.annotation_group
    assert not allele.annotation_splice_variant


def test_annotation_null():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01", annotation="N")
    assert allele.annotation_null
    assert not allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert not allele.annotation_secreted
    assert not allele.annotation_pseudogene
    assert not allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert not allele.annotation_group
    assert not allele.annotation_splice_variant


def test_annotation_cytosolic():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01", annotation="C")
    assert not allele.annotation_null
    assert allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert not allele.annotation_secreted
    assert not allele.annotation_pseudogene
    assert not allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert not allele.annotation_group
    assert not allele.annotation_splice_variant


def test_annotation_secreted():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01", annotation="S")
    assert not allele.annotation_null
    assert not allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert allele.annotation_secreted
    assert not allele.annotation_pseudogene
    assert not allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert not allele.annotation_group
    assert not allele.annotation_splice_variant


def test_annotation_questionable():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01", annotation="Q")
    assert not allele.annotation_null
    assert not allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert not allele.annotation_secreted
    assert not allele.annotation_pseudogene
    assert allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert not allele.annotation_group
    assert not allele.annotation_splice_variant


def test_annotation_group():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01", annotation="G")
    assert not allele.annotation_null
    assert not allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert not allele.annotation_secreted
    assert not allele.annotation_pseudogene
    assert not allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert allele.annotation_group
    assert not allele.annotation_splice_variant


def test_annotation_splice_variant():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01", annotation="Sp")
    assert not allele.annotation_null
    assert not allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert not allele.annotation_secreted
    assert not allele.annotation_pseudogene
    assert not allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert not allele.annotation_group
    assert allele.annotation_splice_variant


def test_annotation_pseudogene():
    allele = Allele.get("HLA", "A", "02", "01", "01", "01", annotation="Ps")
    assert not allele.annotation_null
    assert not allele.annotation_cystosolic
    assert not allele.annotation_aberrant_expression
    assert not allele.annotation_secreted
    assert allele.annotation_pseudogene
    assert not allele.annotation_questionable
    assert not allele.annotation_low_expression
    assert not allele.annotation_group
    assert not allele.annotation_splice_variant


# ---- Field width normalization tests ----


def test_normalize_single_digit_to_two():
    """Single-digit fields are zero-padded to 2 digits for standard HLA genes."""
    allele = parse("A*2:1")
    eq_(list(allele.allele_fields), ["02", "01"])


def test_normalize_already_two_digit_unchanged():
    """Two-digit fields are not changed."""
    allele = parse("A*02:01")
    eq_(list(allele.allele_fields), ["02", "01"])


def test_normalize_three_digit_preserved():
    """Three-digit fields wider than the minimum are never truncated."""
    allele = parse("DPB1*105:01")
    eq_(list(allele.allele_fields), ["105", "01"])


def test_normalize_mica_three_digit():
    """MICA uses 3-digit minimum for first field."""
    allele = parse("MICA*01")
    eq_(list(allele.allele_fields), ["001"])


def test_normalize_mica_already_three_digit():
    """MICA with already 3-digit first field is unchanged."""
    allele = parse("MICA*002:01:01")
    eq_(list(allele.allele_fields), ["002", "01", "01"])


def test_normalize_equality():
    """Alleles with different input widths compare equal after normalization."""
    a1 = parse("A*2:1")
    a2 = parse("A*02:01")
    assert a1 == a2
    assert hash(a1) == hash(a2)


def test_normalize_equality_mica():
    """MICA alleles with different input widths compare equal."""
    a1 = parse("MICA*01")
    a2 = parse("MICA*001")
    assert a1 == a2


def test_normalize_to_string():
    """Normalized alleles produce canonical to_string output."""
    allele = parse("A*2:1")
    eq_(allele.to_string(), "HLA-A*02:01")


def test_normalize_off():
    """normalize_fields=False preserves original widths."""
    allele = Allele.get("HLA", "A", "2", "1", normalize_fields=False)
    eq_(list(allele.allele_fields), ["2", "1"])


def test_normalize_factory_get_with_gene():
    """get_with_gene also normalizes."""
    from mhcgnomes import Gene

    gene = Gene.get("HLA", "A")
    allele = Allele.get_with_gene(gene, ["2", "1"])
    eq_(list(allele.allele_fields), ["02", "01"])



"""Retrieve pinned IMGT/IPD inputs through verified, reusable cache layers."""

from __future__ import annotations

import argparse
import hashlib
import json
import os
import shutil
import sys
import tempfile
import time
import urllib.error
import urllib.parse
import urllib.request
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any

MANIFEST_PATH = Path(__file__).with_name("data") / "external_sources.json"
CACHE_ENV = "MHCGNOMES_DATA_CACHE"
MIRRORS_ENV = "MHCGNOMES_DATA_MIRRORS"
USER_AGENT = "mhcgnomes-external-data/1 (+https://github.com/pirl-unc/mhcgnomes)"


class ExternalDataError(RuntimeError):
    """Raised when an external source cannot be resolved and verified."""


@dataclass(frozen=True)
class Source:
    id: str
    filename: str
    provider: str
    release: str
    url: str
    sha256: str
    license_url: str
    groups: tuple[str, ...]
    default: bool


def default_cache_dir() -> Path:
    configured = os.environ.get(CACHE_ENV)
    if configured:
        return Path(configured).expanduser()
    xdg_cache = os.environ.get("XDG_CACHE_HOME")
    cache_root = Path(xdg_cache).expanduser() if xdg_cache else Path.home() / ".cache"
    return cache_root / "mhcgnomes" / "external-data-v1"


def _validate_filename(filename: str, source_id: str) -> None:
    path = Path(filename)
    if path.is_absolute() or len(path.parts) != 1 or filename in {"", ".", ".."}:
        raise ExternalDataError(f"Source {source_id!r} has unsafe filename {filename!r}")


def _parse_source(raw: dict[str, Any]) -> Source:
    if not isinstance(raw, dict):
        raise ExternalDataError("Manifest source entries must be JSON objects")
    required = {
        "id",
        "filename",
        "provider",
        "release",
        "url",
        "sha256",
        "license_url",
        "groups",
        "default",
    }
    missing = required.difference(raw)
    if missing:
        raise ExternalDataError(f"Manifest source is missing fields: {sorted(missing)}")

    if not isinstance(raw["groups"], list) or not all(
        isinstance(group, str) and group for group in raw["groups"]
    ):
        raise ExternalDataError(f"Source {raw['id']!r} must have a list of non-empty groups")
    if not isinstance(raw["default"], bool):
        raise ExternalDataError(f"Source {raw['id']!r} must have a boolean default value")

    source = Source(
        id=str(raw["id"]),
        filename=str(raw["filename"]),
        provider=str(raw["provider"]),
        release=str(raw["release"]),
        url=str(raw["url"]),
        sha256=str(raw["sha256"]).lower(),
        license_url=str(raw["license_url"]),
        groups=tuple(raw["groups"]),
        default=raw["default"],
    )
    if not source.id:
        raise ExternalDataError("Manifest source IDs must not be empty")
    _validate_filename(source.filename, source.id)
    if len(source.sha256) != 64 or any(c not in "0123456789abcdef" for c in source.sha256):
        raise ExternalDataError(f"Source {source.id!r} has invalid SHA-256")
    if urllib.parse.urlparse(source.url).scheme != "https":
        raise ExternalDataError(f"Source {source.id!r} must use an HTTPS primary URL")
    if urllib.parse.urlparse(source.license_url).scheme != "https":
        raise ExternalDataError(f"Source {source.id!r} must use an HTTPS license URL")
    if not source.groups:
        raise ExternalDataError(f"Source {source.id!r} must belong to at least one group")
    return source


def load_manifest(path: Path = MANIFEST_PATH) -> tuple[Source, ...]:
    try:
        raw = json.loads(path.read_text(encoding="utf-8"))
    except (OSError, json.JSONDecodeError) as exc:
        raise ExternalDataError(f"Cannot read external-data manifest {path}: {exc}") from exc
    if (
        not isinstance(raw, dict)
        or raw.get("schema_version") != 1
        or not isinstance(raw.get("sources"), list)
    ):
        raise ExternalDataError(f"Unsupported external-data manifest schema in {path}")

    sources = tuple(_parse_source(item) for item in raw["sources"])
    ids = [source.id for source in sources]
    filenames = [source.filename for source in sources]
    if len(ids) != len(set(ids)):
        raise ExternalDataError("External-data manifest contains duplicate source IDs")
    if len(filenames) != len(set(filenames)):
        raise ExternalDataError("External-data manifest contains duplicate filenames")
    return sources


def sha256_file(path: Path) -> str:
    digest = hashlib.sha256()
    with path.open("rb") as handle:
        for chunk in iter(lambda: handle.read(1024 * 1024), b""):
            digest.update(chunk)
    return digest.hexdigest()


def is_verified(path: Path, expected_sha256: str) -> bool:
    try:
        return path.is_file() and sha256_file(path) == expected_sha256
    except OSError:
        return False


def _atomic_copy(source_path: Path, destination: Path) -> None:
    destination.parent.mkdir(parents=True, exist_ok=True)
    descriptor, temporary_name = tempfile.mkstemp(
        prefix=f".{destination.name}.", suffix=".tmp", dir=destination.parent
    )
    os.close(descriptor)
    temporary = Path(temporary_name)
    try:
        shutil.copyfile(source_path, temporary)
        os.replace(temporary, destination)
    finally:
        temporary.unlink(missing_ok=True)


def _download_to_cache(url: str, source: Source, cache_path: Path, retries: int = 3) -> None:
    cache_path.parent.mkdir(parents=True, exist_ok=True)
    last_error: Exception | None = None
    for attempt in range(retries):
        descriptor, temporary_name = tempfile.mkstemp(
            prefix=f".{source.sha256}.", suffix=".tmp", dir=cache_path.parent
        )
        os.close(descriptor)
        temporary = Path(temporary_name)
        try:
            request = urllib.request.Request(url, headers={"User-Agent": USER_AGENT})
            digest = hashlib.sha256()
            with (
                urllib.request.urlopen(request, timeout=60) as response,
                temporary.open("wb") as output,
            ):
                while chunk := response.read(1024 * 1024):
                    output.write(chunk)
                    digest.update(chunk)
            actual = digest.hexdigest()
            if actual != source.sha256:
                raise ExternalDataError(
                    f"Checksum mismatch for {source.id}: expected {source.sha256}, got {actual}"
                )
            os.replace(temporary, cache_path)
            return
        except ExternalDataError:
            raise
        except urllib.error.HTTPError as exc:
            last_error = exc
            retryable = exc.code in {408, 425, 429} or exc.code >= 500
            if not retryable:
                break
            if attempt + 1 < retries:
                time.sleep(2**attempt)
        except (OSError, urllib.error.URLError) as exc:
            last_error = exc
            if attempt + 1 < retries:
                time.sleep(2**attempt)
        finally:
            temporary.unlink(missing_ok=True)
    raise ExternalDataError(f"Failed to download {source.id} from {url}: {last_error}")


def _mirror_url(mirror: str, filename: str) -> str:
    return f"{mirror.rstrip('/')}/{urllib.parse.quote(filename)}"


def _try_local_mirror(mirror: Path, source: Source, cache_path: Path) -> bool:
    for candidate in (mirror / source.filename, mirror / source.sha256):
        if is_verified(candidate, source.sha256):
            _atomic_copy(candidate, cache_path)
            return True
    return False


def _is_network_mirror(mirror: str) -> bool:
    return urllib.parse.urlparse(mirror).scheme in {"http", "https"}


def resolve_source(
    source: Source,
    *,
    destination_dir: Path,
    cache_dir: Path,
    mirrors: Sequence[str] = (),
    offline: bool = False,
) -> tuple[Path, str]:
    destination = destination_dir / source.filename
    if is_verified(destination, source.sha256):
        return destination, "destination"

    cache_path = cache_dir / source.sha256
    origin = "cache"
    if not is_verified(cache_path, source.sha256):
        origin = ""
        for mirror in mirrors:
            if _is_network_mirror(mirror):
                if offline:
                    continue
                for mirror_key in (source.filename, source.sha256):
                    try:
                        _download_to_cache(_mirror_url(mirror, mirror_key), source, cache_path)
                    except ExternalDataError:
                        continue
                    origin = f"mirror {mirror}"
                    break
                if origin:
                    break
            if _try_local_mirror(Path(mirror).expanduser(), source, cache_path):
                origin = f"mirror {mirror}"
                break

        if not origin:
            if offline:
                raise ExternalDataError(
                    f"Offline and no verified cache or local mirror contains {source.id}"
                )
            _download_to_cache(source.url, source, cache_path)
            origin = "official source"

    _atomic_copy(cache_path, destination)
    if not is_verified(destination, source.sha256):
        raise ExternalDataError(f"Failed to materialize verified source {source.id}")
    return destination, origin


def environment_mirrors() -> tuple[str, ...]:
    value = os.environ.get(MIRRORS_ENV, "")
    return tuple(part.strip() for part in value.split(";") if part.strip())


def select_sources(
    sources: Sequence[Source],
    *,
    source_ids: Sequence[str] = (),
    groups: Sequence[str] = (),
    all_sources: bool = False,
) -> tuple[Source, ...]:
    known_ids = {source.id for source in sources}
    known_groups = {group for source in sources for group in source.groups}
    unknown_ids = set(source_ids).difference(known_ids)
    unknown_groups = set(groups).difference(known_groups)
    if unknown_ids:
        raise ExternalDataError(f"Unknown source IDs: {sorted(unknown_ids)}")
    if unknown_groups:
        raise ExternalDataError(f"Unknown source groups: {sorted(unknown_groups)}")

    requested_ids = set(source_ids)
    requested_groups = set(groups)
    return tuple(
        source
        for source in sources
        if all_sources
        or (not requested_ids and not requested_groups and source.default)
        or source.id in requested_ids
        or requested_groups.intersection(source.groups)
    )


def materialize_sources(
    sources: Iterable[Source],
    *,
    destination_dir: Path,
    cache_dir: Path,
    mirrors: Sequence[str] = (),
    offline: bool = False,
) -> tuple[Path, ...]:
    paths = []
    for source in sources:
        path, origin = resolve_source(
            source,
            destination_dir=destination_dir,
            cache_dir=cache_dir,
            mirrors=mirrors,
            offline=offline,
        )
        print(f"{source.id}: {path} ({origin})")
        paths.append(path)
    return tuple(paths)


def _add_download_arguments(parser: argparse.ArgumentParser) -> None:
    """Attach source-selection and retrieval options to the download command."""
    parser.add_argument("source_ids", nargs="*", help="Source IDs (defaults to maintained inputs)")
    parser.add_argument("--group", action="append", default=[], help="Select a purpose group")
    parser.add_argument("--all", action="store_true", help="Download every registered source")
    parser.add_argument("--manifest", type=Path, default=MANIFEST_PATH)
    parser.add_argument("--cache-dir", type=Path, default=default_cache_dir())
    parser.add_argument("--destination", type=Path, default=Path(".external-data"))
    parser.add_argument(
        "--mirror",
        action="append",
        default=[],
        help="Local directory or HTTP(S) base URL; repeat for fallback order",
    )
    parser.add_argument("--offline", action="store_true", help="Forbid all network access")


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="mhcgnomes data",
        description="List or download checksum-pinned IMGT/IPD inputs.",
    )
    subparsers = parser.add_subparsers(dest="data_command")
    subparsers.add_parser(
        "list",
        help="List registered sources (default)",
        description="List registered external-data sources and purpose groups.",
    )
    download_parser = subparsers.add_parser(
        "download",
        help="Download and cache verified sources",
        description="Download checksum-pinned inputs through local and CI-friendly caches.",
    )
    _add_download_arguments(download_parser)
    return parser


def print_sources(sources: Sequence[Source]) -> None:
    for source in sources:
        groups = ",".join(source.groups)
        default = " default" if source.default else ""
        print(f"{source.id}\t{source.provider} {source.release}\t{groups}{default}")


def main(argv: Sequence[str] | None = None) -> int:
    parser = build_parser()
    args = parser.parse_args(list(argv) if argv is not None else None)
    try:
        if args.data_command in {None, "list"}:
            print_sources(load_manifest())
            return 0
        sources = load_manifest(args.manifest)
        selected = select_sources(
            sources,
            source_ids=args.source_ids,
            groups=args.group,
            all_sources=args.all,
        )
        if not selected:
            raise ExternalDataError("No external sources selected")
        mirrors = (*args.mirror, *environment_mirrors())
        materialize_sources(
            selected,
            destination_dir=args.destination,
            cache_dir=args.cache_dir,
            mirrors=mirrors,
            offline=args.offline,
        )
        return 0
    except ExternalDataError as exc:
        print(f"ERROR: {exc}", file=sys.stderr)
        return 1


if __name__ == "__main__":
    raise SystemExit(main())



#!/usr/bin/env python3
"""
Evaluate mhcgnomes parse rate and correctness on validation data.

Usage:
    python paper/scripts/evaluate.py paper/validation/*.tsv

Input format (TSV):
    raw_string\texpected_species\tsource

Output: summary statistics + per-source breakdown + error analysis.
"""

import argparse
import csv
import sys
from collections import Counter, defaultdict
from pathlib import Path

import mhcgnomes
from mhcgnomes import parse


def load_validation_data(paths):
    """Load validation TSV files. Columns: raw_string, expected_species, source."""
    records = []
    for path in paths:
        with open(path) as f:
            reader = csv.DictReader(f, delimiter="\t")
            for row in reader:
                records.append(row)
    return records


def evaluate(records):
    """Parse each record and classify the result."""
    results = []
    for rec in records:
        raw = rec["raw_string"].strip()
        expected_species = rec.get("expected_species", "").strip()
        source = rec.get("source", "unknown").strip()

        try:
            result = parse(raw, raise_on_error=True)
            parsed = True
            result_type = type(result).__name__
            parsed_species = getattr(result, "species", None)
            species_name = parsed_species.name if parsed_species else None
            normalized = result.to_string() if hasattr(result, "to_string") else str(result)

            # Check species correctness if expected_species provided
            if expected_species and species_name:
                species_correct = species_name == expected_species
            else:
                species_correct = None
        except Exception:
            parsed = False
            result_type = None
            species_name = None
            normalized = None
            species_correct = None

        results.append({
            "raw": raw,
            "source": source,
            "expected_species": expected_species,
            "parsed": parsed,
            "result_type": result_type,
            "species_name": species_name,
            "normalized": normalized,
            "species_correct": species_correct,
        })

    return results


def round_trip_check(results):
    """Check that parse(x).to_string() round-trips correctly."""
    rt_pass = 0
    rt_fail = 0
    rt_failures = []

    for r in results:
        if not r["parsed"] or r["normalized"] is None:
            continue
        try:
            reparsed = parse(r["normalized"], raise_on_error=True)
            if reparsed.to_string() == r["normalized"]:
                rt_pass += 1
            else:
                rt_fail += 1
                rt_failures.append((r["normalized"], reparsed.to_string()))
        except Exception:
            rt_fail += 1
            rt_failures.append((r["normalized"], "PARSE_ERROR"))

    return rt_pass, rt_fail, rt_failures


def summarize(results):
    """Print summary statistics."""
    total = len(results)
    parsed = sum(1 for r in results if r["parsed"])
    failed = total - parsed

    print(f"{'Total strings':.<40} {total}")
    print(f"{'Parsed successfully':.<40} {parsed} ({100*parsed/total:.1f}%)")
    print(f"{'Failed to parse':.<40} {failed} ({100*failed/total:.1f}%)")
    print()

    # Species correctness (where expected_species was provided)
    with_expected = [r for r in results if r["expected_species"] and r["parsed"]]
    if with_expected:
        correct = sum(1 for r in with_expected if r["species_correct"])
        wrong = len(with_expected) - correct
        print(f"{'Species correctness (of parsed with expected)':.<40}")
        print(f"  {'Correct':.<38} {correct} ({100*correct/len(with_expected):.1f}%)")
        print(f"  {'Wrong species':.<38} {wrong}")
        print()

    # Normalization yield
    raw_unique = len(set(r["raw"] for r in results if r["parsed"]))
    norm_unique = len(set(r["normalized"] for r in results if r["parsed"]))
    if raw_unique > 0:
        print(f"{'Unique raw strings (parsed)':.<40} {raw_unique}")
        print(f"{'Unique normalized forms':.<40} {norm_unique}")
        print(f"{'Compression ratio':.<40} {raw_unique/norm_unique:.2f}x")
        print()

    # Round-trip fidelity
    rt_pass, rt_fail, rt_failures = round_trip_check(results)
    if rt_pass + rt_fail > 0:
        print(f"{'Round-trip fidelity':.<40}")
        print(f"  {'Pass':.<38} {rt_pass}")
        print(f"  {'Fail':.<38} {rt_fail}")
        if rt_failures:
            print(f"  First 5 failures:")
            for orig, reparsed in rt_failures[:5]:
                print(f"    {orig} → {reparsed}")
        print()

    # Per-source breakdown
    by_source = defaultdict(lambda: {"total": 0, "parsed": 0})
    for r in results:
        by_source[r["source"]]["total"] += 1
        if r["parsed"]:
            by_source[r["source"]]["parsed"] += 1

    print("Per-source parse rates:")
    print(f"  {'Source':<30} {'Parsed':>8} {'Total':>8} {'Rate':>8}")
    for source in sorted(by_source):
        s = by_source[source]
        rate = 100 * s["parsed"] / s["total"] if s["total"] else 0
        print(f"  {source:<30} {s['parsed']:>8} {s['total']:>8} {rate:>7.1f}%")
    print()

    # Result type distribution
    type_counts = Counter(r["result_type"] for r in results if r["parsed"])
    print("Result type distribution:")
    for rtype, count in type_counts.most_common():
        print(f"  {rtype:<30} {count}")
    print()

    # Failure analysis: sample of unparsed strings
    failures = [r["raw"] for r in results if not r["parsed"]]
    if failures:
        print(f"Sample of failed strings ({min(20, len(failures))} of {len(failures)}):")
        for s in failures[:20]:
            print(f"  {s}")


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("files", nargs="+", help="Validation TSV files")
    args = parser.parse_args()

    records = load_validation_data(args.files)
    if not records:
        print("No records loaded.", file=sys.stderr)
        return 1

    results = evaluate(records)
    summarize(results)
    return 0


if __name__ == "__main__":
    sys.exit(main())



import pytest

from mhcgnomes import Gene, ParseError, infer_mhc_class, parse, parse_gene_class

from .common import eq_


def test_infer_mhc_class_from_species_aware_suffix():
    eq_(infer_mhc_class("F10", species="Pogona vitticeps"), "I")
    eq_(infer_mhc_class("Povi-F10"), "I")


@pytest.mark.parametrize(
    "raw_string, expected_class, expected_chain",
    [
        ("RT1.A(u)", "Ia", "alpha"),
        ("RT1.A1(f)", "Ia", "alpha"),
        # RT1.B resolves to Gene(Bb) via gene alias (NCBI GeneID 24738);
        # Bb is specifically a class IIa beta chain gene
        ("RT1.B", "IIa", "beta"),
        ("RT1.D", "II", None),
        ("RT1-Ba", "IIa", "alpha"),
        ("RT1-DOa", "IIb", "alpha"),
        ("RT1-DOb", "IIb", "beta"),
        ("H2-Oa", "IIb", "alpha"),
        ("H2-Ob", "IIb", "beta"),
        ("I-A-beta", "IIa", "beta"),
        ("MHC II H2-IE-beta", "IIa", "beta"),
        ("rt1-El", "I", "alpha"),
        ("rt1-Eg", "I", "alpha"),
    ],
)
def test_infer_mhc_class_handles_rt1_and_h2_edge_cases(raw_string, expected_class, expected_chain):
    eq_(infer_mhc_class(raw_string), expected_class)
    result = parse_gene_class(raw_string)
    eq_(result.mhc_class, expected_class)
    eq_(result.chain, expected_chain)


def test_parse_gene_class_preserves_strict_parse_for_canonical_gene():
    result = parse_gene_class("PogoVitt-DRA")
    eq_(result.species.name, "Pogona vitticeps")
    eq_(result.gene_name, "DRA")
    eq_(result.mhc_class, "IIa")
    eq_(result.chain, "alpha")
    assert not result.non_mhc
    eq_(result.source, "parsed_gene")


def test_parse_gene_class_heuristically_recovers_f10_with_species_hint():
    result = parse_gene_class("F10", species="Pogona vitticeps")
    eq_(result.species.name, "Pogona vitticeps")
    eq_(result.gene_name, "F10")
    eq_(result.mhc_class, "I")
    eq_(result.chain, "alpha")
    assert not result.non_mhc
    eq_(result.source, "heuristic_suffix")


def test_parse_gene_class_heuristically_recovers_dr_1():
    result = parse_gene_class("DR-1", species="Pogona vitticeps")
    eq_(result.gene_name, "DR-1")
    eq_(result.mhc_class, "IIa")
    eq_(result.chain, "alpha")


def test_parse_gene_class_heuristically_recovers_drb_range():
    result = parse_gene_class("DRB1-4", species="Pipra filicauda")
    eq_(result.gene_name, "DRB1-4")
    eq_(result.mhc_class, "IIa")
    eq_(result.chain, "beta")


def test_parse_gene_class_respects_species_context_for_colliding_prefixes():
    fish = parse_gene_class("Hymo-DAB", species="Hypophthalmichthys molitrix")
    gibbon = parse_gene_class("Hymo-DRB1*04:01", species="Hylobates moloch")
    medaka = parse_gene_class("ORLA-UAA", species="Oryzias latipes")

    eq_(fish.species.name, "Hypophthalmichthys molitrix")
    eq_(fish.gene_name, "DAB")
    eq_(fish.chain, "beta")

    eq_(gibbon.species.name, "Hylobates moloch")
    eq_(gibbon.gene_name, "DRB1")
    eq_(gibbon.chain, "beta")

    eq_(medaka.species.name, "Oryzias latipes")
    eq_(medaka.gene_name, "UAA")
    eq_(medaka.chain, "alpha")


def test_parse_gene_class_marks_known_other_genes_as_non_mhc():
    result = parse_gene_class("Tap1", species="Pogona vitticeps")
    eq_(result.gene_name, "TAP1")
    eq_(result.mhc_class, "other")
    eq_(result.chain, None)
    assert result.non_mhc
    assert result.source in {"parsed_gene", "canonical_gene"}


@pytest.mark.parametrize(
    "raw_string, expected_species",
    [
        ("SLA-TAP*1*01:01", "Sus sp."),
        ("SLA-TAP", "Sus sp."),
        ("HLA-TAP", "Homo sapiens"),
        ("Mamu-TAP", "Macaca mulatta"),
        ("Xetr-TAP", "Xenopus tropicalis"),
    ],
)
def test_parse_gene_class_recognizes_ontology_backed_tap_family(raw_string, expected_species):
    result = parse_gene_class(raw_string)

    eq_(result.species.name, expected_species)
    eq_(result.gene_name, "TAP")
    eq_(result.mhc_class, "other")
    eq_(result.chain, None)
    assert result.non_mhc
    eq_(result.source, "ontology_family")


@pytest.mark.parametrize("raw_string", ["SLA-TAP", "Mamu-TAP", "Xetr-TAP"])
def test_strict_parse_does_not_promote_tap_family_to_a_gene(raw_string):
    assert parse(raw_string, raise_on_error=False) is None


@pytest.mark.parametrize("raw_string", ["SLA-TAP0", "SLA-TAPA", "SLA-TAP12"])
def test_parse_gene_class_requires_exact_ontology_family_name(raw_string):
    assert parse_gene_class(raw_string, raise_on_error=False) is None


def test_parse_gene_class_marks_known_non_ontology_helper_gene_as_non_mhc():
    for raw_string, expected_gene in [
        ("Ciita", "CIITA"),
        ("Hm13", "HM13"),
        ("Prr3", "PRR3"),
    ]:
        result = parse_gene_class(raw_string, species="Pogona vitticeps")
        eq_(result.gene_name, expected_gene)
        eq_(result.mhc_class, "other")
        eq_(result.chain, None)
        assert result.non_mhc
        eq_(result.source, "non_mhc_name")


def test_gene_class_info_string_and_record_helpers():
    heuristic = parse_gene_class("F10", species="Pogona vitticeps")
    eq_(heuristic.to_string(), "PogoVitt-F10")
    eq_(heuristic.compact_string(), "F10")
    eq_(heuristic.to_record()["gene_name"], "F10")
    eq_(heuristic.to_record()["mhc_class"], "I")
    eq_(heuristic.to_record()["chain"], "alpha")

    non_mhc = parse_gene_class("Ciita", species="Pogona vitticeps")
    eq_(non_mhc.to_string(include_species=False), "CIITA")
    eq_(non_mhc.to_record()["non_mhc"], True)
    eq_(non_mhc.to_record()["source"], "non_mhc_name")


def test_parse_gene_class_requires_species_context_for_heuristics():
    assert infer_mhc_class("DR-1") is None
    with pytest.raises(ParseError):
        parse_gene_class("DR-1")


def test_parse_remains_conservative_for_gap_report_suffixes():
    for raw in ["Povi-F10", "Miun-E-S", "Gese-DR-1"]:
        assert parse(raw, raise_on_error=False) is None


def test_parse_verbose_mouse_class2_beta_notation():
    result = parse("MHC II H2-IE-beta", raise_on_error=True)
    assert isinstance(result, Gene)
    eq_(result.species.prefix, "H2")
    eq_(result.name, "EB")



"""
Tests for v3.15.0 additions: new species, genes, prefixes,
cetacean class II inheritance, and trailing-MHC prefix stripping.
"""

import pytest

from mhcgnomes import Species, parse

from .common import eq_

# ---------------------------------------------------------------------------
# Trailing-MHC prefix stripping (general mechanism)
# ---------------------------------------------------------------------------


class TestTrailingMhcStrip:
    """ManaMHC-DAB*01 style: trailing MHC suffix on species prefix."""

    def test_mana_mhc_allele(self):
        result = parse("ManaMHC-DAB*01", raise_on_error=True)
        eq_(result.species.name, "Manacus sp.")
        eq_(result.gene.name, "DAB")

    def test_mana_mhc_gene(self):
        result = parse("ManaMHC-UA", raise_on_error=True)
        eq_(result.species.name, "Manacus sp.")
        eq_(result.name, "UA")

    def test_mana_mhc_case_insensitive(self):
        result = parse("ManaMhc-DAB*01", raise_on_error=True)
        eq_(result.species.name, "Manacus sp.")

    def test_trailing_mhc_does_not_break_normal_parsing(self):
        """HLA-A*02:01 should NOT be affected by trailing MHC strip."""
        result = parse("HLA-A*02:01", raise_on_error=True)
        eq_(result.species.name, "Homo sapiens")

    def test_trailing_mhc_does_not_break_leading_mhc(self):
        """MhcTyal-DAB1*01:01 still works (leading Mhc strip)."""
        result = parse("MhcTyal-DAB1*01:01", raise_on_error=True)
        eq_(result.species.prefix, "TytoAlba")

    def test_bare_mana_still_works(self):
        """Mana-DAB*01 (without MHC suffix) should still parse."""
        result = parse("Mana-DAB*01", raise_on_error=True)
        eq_(result.species.name, "Manacus sp.")


# ---------------------------------------------------------------------------
# New species
# ---------------------------------------------------------------------------


class TestNewSpeciesV315:
    """Species added in v3.15.0."""

    @pytest.mark.parametrize(
        "prefix, latin",
        [
            ("Cape", "Caperea marginata"),
            ("Rhsi", "Rhinopithecus strykeri"),
            ("Mygl", "Myodes glareolus"),
            ("Clgl", "Myodes glareolus"),  # synonym prefix
            ("Odhe", "Odocoileus hemionus"),
            ("Loaf", "Loxodonta africana"),
            ("Nesc", "Neomonachus schauinslandi"),
            ("Dele", "Delphinapterus leucas"),
            ("Oror", "Orcinus orca"),
            ("Phca", "Physeter macrocephalus"),
            ("Igig", "Iguana iguana"),
        ],
    )
    def test_new_species_lookup(self, prefix, latin):
        sp = Species.get(prefix)
        assert sp is not None, f"Species.get({prefix!r}) returned None"
        eq_(sp.name, latin)

    def test_manacus_species(self):
        sp = Species.get("Mana")
        assert sp is not None
        eq_(sp.name, "Manacus sp.")

    def test_myodes_synonym(self):
        """Clethrionomys glareolus is a synonym for Myodes glareolus."""
        sp = Species.get("Clethrionomys glareolus")
        assert sp is not None
        eq_(sp.name, "Myodes glareolus")


# ---------------------------------------------------------------------------
# Missing prefixes for existing species
# ---------------------------------------------------------------------------


class TestMissingPrefixes:
    """Alternative 3-letter prefixes for primates and other species."""

    def test_ppa_bonobo(self):
        sp = Species.get("Ppa")
        assert sp is not None
        eq_(sp.name, "Pan paniscus")

    def test_ptr_chimpanzee(self):
        sp = Species.get("Ptr")
        assert sp is not None
        eq_(sp.name, "Pan troglodytes")

    def test_gbb_gorilla(self):
        sp = Species.get("Gbb")
        assert sp is not None
        eq_(sp.name, "Gorilla beringei")

    def test_gbg_gorilla(self):
        sp = Species.get("Gbg")
        assert sp is not None
        eq_(sp.name, "Gorilla beringei")

    def test_ggg_gorilla(self):
        sp = Species.get("Ggg")
        assert sp is not None
        eq_(sp.name, "Gorilla gorilla")

    def test_caur_goldfish(self):
        sp = Species.get("Caur")
        assert sp is not None
        eq_(sp.name, "Carassius auratus")


# ---------------------------------------------------------------------------
# Cetacean class II inheritance
# ---------------------------------------------------------------------------


class TestCetaceaClassII:
    """All cetaceans inherit mammalian class II from Cetacea sp."""

    @pytest.mark.parametrize(
        "gene",
        ["DQA1", "DQB1", "DRA", "DRB", "DRB1", "DPA1", "DPB1", "DOA", "DOB", "DMA", "DMB"],
    )
    def test_cetacea_clade_has_gene(self, gene):
        sp = Species.get("Cetacea sp.")
        assert sp.find_matching_gene_name(gene) is not None, gene

    def test_dolphin_inherits_dqa1(self):
        sp = Species.get("Tutr")
        assert sp.find_matching_gene_name("DQA1") is not None

    def test_dolphin_inherits_dqb1(self):
        sp = Species.get("Tutr")
        assert sp.find_matching_gene_name("DQB1") is not None

    def test_dolphin_inherits_doa(self):
        sp = Species.get("Tutr")
        assert sp.find_matching_gene_name("DOA") is not None

    def test_orca_inherits_class_ii(self):
        sp = Species.get("Orcinus orca")
        assert sp.find_matching_gene_name("DQA1") is not None
        assert sp.find_matching_gene_name("DPB1") is not None

    def test_beluga_inherits_class_ii(self):
        sp = Species.get("Delphinapterus leucas")
        assert sp.find_matching_gene_name("DRB1") is not None

    def test_parse_tutr_dqb1_allele(self):
        result = parse("Tutr-DQB1*01:01", raise_on_error=True)
        eq_(result.species.name, "Tursiops truncatus")


# ---------------------------------------------------------------------------
# Missing genes on existing species
# ---------------------------------------------------------------------------


class TestMissingGenesV315:
    """Genes added to existing species."""

    def test_salmon_uga(self):
        sp = Species.get("Sasa")
        assert sp.find_matching_gene_name("UGA") is not None

    def test_salmon_uha(self):
        sp = Species.get("Sasa")
        assert sp.find_matching_gene_name("UHA") is not None

    def test_salmon_ze(self):
        sp = Species.get("Sasa")
        assert sp.find_matching_gene_name("ZE") is not None

    def test_iguana_ub(self):
        sp = Species.get("AmblCris")
        assert sp.find_matching_gene_name("UB") is not None

    def test_koala_uh(self):
        sp = Species.get("PhasCine")
        assert sp.find_matching_gene_name("UH") is not None

    def test_equus_dqa_family(self):
        sp = Species.get("ELA")
        assert sp.find_matching_gene_name("DQA") is not None

    def test_equus_dqb_family(self):
        sp = Species.get("ELA")
        assert sp.find_matching_gene_name("DQB") is not None

    def test_equus_drb_family(self):
        sp = Species.get("ELA")
        assert sp.find_matching_gene_name("DRB") is not None



import pytest

from mhcgnomes import Serotype, Species, parse

from .common import eq_


def test_HLA_A2_in_species_serotype_dictionary():
    human = Species.get("HLA")
    assert human is not None
    assert "A2" in human.serotypes


def test_parse_HLA_A2_serotype():
    result = parse("HLA-A2")
    assert result is not None
    eq_(result.name, "A2")
    eq_(result.species.prefix, "HLA")


# ---- IEF/CTL dot-notation subtypes ----


@pytest.mark.parametrize(
    "name, expected_alleles",
    [
        # A2 IEF subtypes (van der Poel et al. 1983)
        ("A2.1", ["A*02:01"]),
        ("A2.2F", ["A*02:02"]),
        ("A2.2Y", ["A*02:05"]),
        ("A2.3", ["A*02:03"]),
        ("A2.4", ["A*02:04"]),
        ("A2.4a", ["A*02:06"]),
        ("A2.4b", ["A*02:07"]),
        ("A2.5", ["A*02:11"]),
        # B27 IEF subtypes (Rojo et al. 1985)
        ("B27.1", ["B*27:05"]),
        ("B27.2", ["B*27:02"]),
        ("B27.3", ["B*27:04"]),
        # B35 IEF subtypes (Yang et al. 1995)
        ("B35.1", ["B*35:08"]),
        ("B35.2", ["B*35:02", "B*35:04"]),
        ("B35.3", ["B*35:01", "B*35:03"]),
        # B7 IEF subtype (Grumet et al. 1995)
        ("B7.1", ["B*07:05"]),
    ],
)
def test_ief_dot_notation_serotypes(name, expected_alleles):
    """IEF/CTL dot-notation subtypes parse as serotypes with correct alleles."""
    result = parse(name)
    assert isinstance(result, Serotype), (
        f"Expected Serotype for '{name}' but got {type(result).__name__}: {result}"
    )
    eq_(result.name, name)
    actual_alleles = sorted(a.to_string(include_species=False) for a in result.alleles)
    eq_(actual_alleles, sorted(expected_alleles))



from pathlib import Path

import pytest
import yaml

from mhcgnomes import Allele, Gene, Species, parse

from .common import eq_


def load_raw_species_yaml():
    species_path = Path(__file__).parent.parent / "mhcgnomes" / "data" / "species.yaml"
    with species_path.open() as fd:
        return yaml.safe_load(fd)


TARGETED_REDUNDANT_DIRECT_GENE_BLOCKS_REMOVED = [
    "Balaenoptera musculus",
    "Balaenoptera acutorostrata",
    "Balaenoptera edeni",
    "Cephalorhynchus hectori",
    "Delphinus delphis",
    "Eubalaena australis",
    "Globicephala melas",
    "Grampus griseus",
    "Kogia breviceps",
    "Lagenorhynchus obscurus",
    "Mesoplodon densirostris",
    "Mesoplodon europaeus",
    "Mesoplodon grayi",
    "Megaptera novaeangliae",
    "Steno bredanensis",
    "Stenella coeruleoalba",
    "Tursiops truncatus",
    "Ziphius cavirostris",
    "Strigidae sp.",
    "Tyto alba",
    "Macaca silenus",
    "Macaca thibetana",
    "Falco sp.",
    "Accipitriformes sp.",
    "Gruiformes sp.",
    "Ardeidae sp.",
]


def test_targeted_subset_only_direct_gene_blocks_removed_from_raw_yaml():
    species_data = load_raw_species_yaml()
    for latin_name in TARGETED_REDUNDANT_DIRECT_GENE_BLOCKS_REMOVED:
        assert "genes" not in species_data[latin_name], latin_name


def test_goat_inherits_helper_genes_from_root():
    """Capra sp. no longer declares TAP1/TAP2/TAPBP/B2M directly;
    they are inherited from the Gnathostomata sp. root."""
    species_data = load_raw_species_yaml()
    goat_genes = species_data["Capra sp."]["genes"]
    assert "other" not in goat_genes
    # Verify they're still on the root
    root_genes = species_data["Gnathostomata sp."]["genes"]
    assert "other" in root_genes
    assert {"TAP1", "TAP2", "TAPBP", "B2M"}.issubset(set(root_genes["other"]))


def test_human_and_rabbit_do_not_use_generic_top_level_class_ii_bucket():
    species_data = load_raw_species_yaml()
    assert "II" not in species_data["Homo sapiens"]["genes"]
    assert "II" not in species_data["Oryctolagus sp."]["genes"]


def test_rat_a3_only_appears_in_class_ia_in_raw_yaml():
    species_data = load_raw_species_yaml()
    rat_genes = species_data["Rattus sp."]["genes"]
    assert "A3" in rat_genes["Ia"]
    assert "A3" not in rat_genes["I"]


@pytest.mark.parametrize(
    "species_prefix,gene_name",
    [
        ("Bamu", "DQA"),
        ("Baac", "S"),
        ("Tyal", "DAB1"),
        ("Stoc", "DAB2"),
        ("Masi", "DRB4"),
        ("Math", "DPB1"),
    ],
)
def test_targeted_cleanup_species_still_inherit_representative_genes(species_prefix, gene_name):
    gene = Gene.get(species_prefix, gene_name)
    assert gene is not None, f"Gene.get({species_prefix!r}, {gene_name!r}) returned None"
    eq_(parse(f"{species_prefix}-{gene_name}", raise_on_error=True), gene)


@pytest.mark.parametrize(
    "species_prefix,gene_name,expected_class",
    [
        ("CLA", "TAP1", "other"),
        ("CLA", "B2M", "other"),
        ("HLA", "DRB6", "IIa"),
        ("RLA", "DPA", "IIa"),
        ("RT1", "A3", "Ia"),
    ],
)
def test_runtime_gene_classes_match_cleaned_ontology(species_prefix, gene_name, expected_class):
    species = Species.get(species_prefix)
    eq_(species.get_mhc_class_of_gene(gene_name), expected_class)


@pytest.mark.parametrize("raw", ["Ctid-UHA103", "Dila-a1", "Dila-a30"])
def test_sequence_like_labels_remain_rejected_as_canonical_genes(raw):
    assert parse(raw, raise_on_error=False) is None


def test_pelodiscus_numbered_b_label_parses_as_b_allele():
    expected = Allele.get("Pesi", "B", "01")
    assert expected is not None
    eq_(parse("Pesi-B01", raise_on_error=True), expected)



#!/usr/bin/env python3
"""
Collect MHC allele names from GenBank feature annotations.

Queries NCBI for MHC sequences and extracts the /allele= qualifier,
/gene= qualifier, and organism metadata. These represent real-world
MHC nomenclature as submitted by individual researchers.

Usage:
    python paper/scripts/collect_genbank.py --query "MHC[Gene] AND Aves[Organism]" --max 500
    python paper/scripts/collect_genbank.py --query "MHC class I[Title] AND fish" --max 500
"""

import argparse
import csv
import sys
import time
from io import StringIO

try:
    from Bio import Entrez, SeqIO
except ImportError:
    print("Requires biopython: pip install biopython", file=sys.stderr)
    sys.exit(1)

Entrez.email = "mhcgnomes-paper@example.com"


def search_genbank(query, max_results=500):
    """Search GenBank and return accession IDs."""
    handle = Entrez.esearch(db="nucleotide", term=query, retmax=max_results)
    result = Entrez.read(handle)
    handle.close()
    return result["IdList"]


def fetch_records(ids, batch_size=20):
    """Fetch GenBank records in batches."""
    records = []
    for i in range(0, len(ids), batch_size):
        batch = ids[i : i + batch_size]
        handle = Entrez.efetch(db="nucleotide", id=batch, rettype="gb", retmode="text")
        text = handle.read()
        handle.close()
        for record in SeqIO.parse(StringIO(text), "genbank"):
            records.append(record)
        time.sleep(0.5)  # be polite to NCBI
    return records


def extract_mhc_names(records):
    """Extract MHC allele/gene names from GenBank records."""
    rows = []
    for record in records:
        organism = record.annotations.get("organism", "")
        accession = record.id

        for feature in record.features:
            allele = feature.qualifiers.get("allele", [None])[0]
            gene = feature.qualifiers.get("gene", [None])[0]
            product = feature.qualifiers.get("product", [None])[0]
            note = feature.qualifiers.get("note", [None])[0]

            # Collect any MHC-related name
            for name_source, name in [
                ("allele", allele),
                ("gene", gene),
                ("product", product),
            ]:
                if name and any(
                    kw in name.upper()
                    for kw in [
                        "MHC",
                        "HLA",
                        "H-2",
                        "H2-",
                        "RT1",
                        "CLASS I",
                        "CLASS II",
                        "HISTOCOMPATIBILITY",
                        "DRB",
                        "DQB",
                        "DPB",
                        "DAB",
                        "UAA",
                        "UBA",
                        "BLB",
                        "BF",
                        "TAP",
                        "B2M",
                    ]
                ):
                    rows.append(
                        {
                            "raw_string": name,
                            "expected_species": organism,
                            "source": f"genbank:{accession}",
                            "qualifier": name_source,
                        }
                    )

    return rows


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--query", required=True, help="GenBank search query")
    parser.add_argument("--max", type=int, default=500, help="Max results")
    parser.add_argument("--output", default="-", help="Output TSV file")
    args = parser.parse_args()

    print(f"Searching GenBank: {args.query}", file=sys.stderr)
    ids = search_genbank(args.query, args.max)
    print(f"Found {len(ids)} records", file=sys.stderr)

    if not ids:
        return

    print("Fetching records...", file=sys.stderr)
    records = fetch_records(ids)
    print(f"Fetched {len(records)} records", file=sys.stderr)

    rows = extract_mhc_names(records)
    print(f"Extracted {len(rows)} MHC names", file=sys.stderr)

    out = sys.stdout if args.output == "-" else open(args.output, "w")
    writer = csv.DictWriter(
        out,
        fieldnames=["raw_string", "expected_species", "source", "qualifier"],
        delimiter="\t",
    )
    writer.writeheader()
    writer.writerows(rows)

    if args.output != "-":
        out.close()


if __name__ == "__main__":
    main()



"""
Tests for batch-added species from the mhcseqs P2/P5 requests.
Covers raptors, cranes, passerines, additional fish, amphibians,
sharks, tuatara, and taxonomy reclassifications.
"""

import pytest

from mhcgnomes import Gene, Species, parse

from .common import eq_

BATCH_SPECIES = [
    # Raptors
    ("Fape", "Falco peregrinus"),
    ("Fati", "Falco tinnunculus"),
    ("Aqch", "Aquila chrysaetos"),
    ("ButeoBute", "Buteo buteo"),
    ("Ciae", "Circus aeruginosus"),
    ("Gyfu", "Gyps fulvus"),
    # Cranes
    ("Grja", "Grus japonensis"),
    ("Anan", "Antigone antigone"),
    ("Bare", "Balearica regulorum"),
    # Passerines
    ("Pado", "Passer domesticus"),
    ("Coco", "Corvus cornix"),
    ("Tume", "Turdus merula"),
    ("Laco", "Lanius collurio"),
    ("Emci", "Emberiza citrinella"),
    # Shorebirds
    ("Anni", "Anarhynchus nivosus"),
    # Fish
    ("Gaac", "Gasterosteus aculeatus"),
    ("Icpu", "Ictalurus punctatus"),
    ("Hiab", "Hippocampus abdominalis"),
    ("Caau", "Carassius auratus"),
    ("Clma", "Clarias magur"),
    # Amphibians
    ("Lica", "Lithobates catesbeianus"),
    ("Licl", "Lithobates clamitans"),
    ("Amme", "Ambystoma mexicanum"),
    ("Amti", "Ambystoma tigrinum"),
    # Tuatara
    ("Sppu", "Sphenodon punctatus"),
    # Marsupial
    ("Noeu", "Notamacropus eugenii"),
    # Sharks
    ("Gici", "Ginglymostoma cirratum"),
    ("Hefr", "Heterodontus francisci"),
    ("Sqac", "Squalus acanthias"),
    ("Trsc", "Triakis scyllium"),
]


@pytest.mark.parametrize("prefix,latin", BATCH_SPECIES)
def test_batch_species_parseable(prefix, latin):
    species = Species.get(prefix)
    assert species is not None, f"Species.get({prefix!r}) returned None"
    eq_(species.name, latin)
    eq_(parse(prefix, raise_on_error=True), species)
    # Latin name should also work
    eq_(Species.get(latin), species)


# Taxonomy reclassification aliases
RECLASSIFICATION_ALIASES = [
    ("Gran", "Antigone antigone"),  # Grus → Antigone
    ("Chni", "Anarhynchus nivosus"),  # Charadrius → Anarhynchus
    ("Raca", "Lithobates catesbeianus"),  # Rana → Lithobates
    ("Racl", "Lithobates clamitans"),  # Rana → Lithobates
    ("Maeu", "Notamacropus eugenii"),  # Macropus → Notamacropus
]


@pytest.mark.parametrize("old_prefix,latin", RECLASSIFICATION_ALIASES)
def test_old_genus_prefix_resolves_to_current_species(old_prefix, latin):
    species = Species.get(old_prefix)
    assert species is not None, f"Old prefix {old_prefix!r} did not resolve"
    eq_(species.name, latin)


GAP_REPORT_SPECIES = [
    ("Anda", "Andrias davidianus", "UAA"),
    ("Geja", "Gekko japonicus", "DAA"),
    ("Spau", "Sparus aurata", "UAA"),
    ("Laca", "Lates calcarifer", "UAA"),
    ("Safa", "Salarias fasciatus", "UAA"),
    ("Acda", "Acipenser dabryanus", "UAA"),
    ("Pybi", "Python bivittatus", "DRA"),
    ("Poli", "Podarcis lilfordi", "DAA"),
    ("Miun", "Microcaecilia unicolor", "DAA"),
    ("Scsc", "Scomber scombrus", "UAA"),
    ("Stoc", "Strix occidentalis caurina", "DAB1"),
]


@pytest.mark.parametrize("prefix,latin,gene_name", GAP_REPORT_SPECIES)
def test_gap_report_species_parseable_with_example_gene(prefix, latin, gene_name):
    species = Species.get(prefix)
    assert species is not None, f"Species.get({prefix!r}) returned None"
    eq_(species.name, latin)
    eq_(Species.get(latin), species)

    gene = Gene.get(prefix, gene_name)
    assert gene is not None, f"Gene.get({prefix!r}, {gene_name!r}) returned None"
    eq_(parse(f"{prefix}-{gene_name}", raise_on_error=True), gene)


PROMOTED_PUBLISHED_PREFIX_ALIASES = [
    ("AndrDavi", "Anda", "UAA"),
    ("SparAura", "Spau", "UAA"),
    ("AcipDabr", "Acda", "UAA"),
]


@pytest.mark.parametrize(
    "legacy_prefix,canonical_prefix,gene_name", PROMOTED_PUBLISHED_PREFIX_ALIASES
)
def test_promoted_published_prefixes_keep_legacy_aliases(
    legacy_prefix, canonical_prefix, gene_name
):
    legacy_species = Species.get(legacy_prefix)
    canonical_species = Species.get(canonical_prefix)

    assert legacy_species is not None
    assert canonical_species is not None
    eq_(legacy_species, canonical_species)

    gene = Gene.get(canonical_prefix, gene_name)
    assert gene is not None
    eq_(parse(f"{legacy_prefix}-{gene_name}", raise_on_error=True), gene)


@pytest.mark.parametrize(
    "species_prefix,gene_name",
    [
        ("Fape", "DAA"),
        ("Fape", "DRA"),
        ("Aqch", "DRB"),
        ("Grja", "DAB"),
        ("Egga", "DAB1"),
        ("Egga", "DAB2"),
    ],
)
def test_cleaned_bird_clades_still_inherit_runtime_genes(species_prefix, gene_name):
    gene = Gene.get(species_prefix, gene_name)
    assert gene is not None, f"Gene.get({species_prefix!r}, {gene_name!r}) returned None"
    eq_(parse(f"{species_prefix}-{gene_name}", raise_on_error=True), gene)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps


def ok_(a, s=None):
    if s is None:
        assert a
    else:
        assert a, s


def eq_(a, b, s=None):
    if s is None:
        assert a == b
    else:
        assert a == b, s


def neq_(a, b, s=None):
    if s is None:
        assert a != b
    else:
        assert a != b, s


def gt_(a, b, s=None):
    if s is None:
        assert a > b
    else:
        assert a > b, s


def lt_(a, b, s=None):
    if s is None:
        assert a < b
    else:
        assert a < b, s


def gte_(a, b, s=None):
    if s is None:
        assert a >= b
    else:
        assert a >= b, s


def lte_(a, b, s=None):
    if s is None:
        assert a <= b
    else:
        assert a <= b, s


def almost_eq_(a, b, tol=1e-6, s=None):
    if s is None:
        assert abs(a - b) < tol
    else:
        assert abs(a - b) < tol, s


class assert_raises:
    def __init__(self, *exception_types):
        self.exception_types = exception_types

    def __enter__(self):
        pass

    def to_string(self):
        return " or ".join([f"{e}" for e in self.exception_types])

    def __exit__(self, type, value, traceback):
        if type is None:
            raise AssertionError(f"Expected exception {self.to_string()} not raised")
        if type not in self.exception_types:
            raise AssertionError(f"Expected exception {self.to_string()}, got {type}")
        return True


def raises(*exception_types):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            with assert_raises(*exception_types):
                f(*args, **kwargs)

        return wrapper

    return decorator



"""
Tests for Species identity model where latin name is the canonical identity
and prefixes/common names are aliases that may be ambiguous.
"""

from collections import defaultdict

import pytest

from mhcgnomes import Allele, Gene, ParseError, Species, parse
from mhcgnomes.common import normalize_string
from mhcgnomes.data import species as species_data
from mhcgnomes.species import _make_5_5_prefix, _make_long_prefix, raw_species_dict

from .common import eq_

# ---------------------------------------------------------------------------
# 1. Species identity via latin name
# ---------------------------------------------------------------------------


def test_species_get_by_latin_name_human():
    species = Species.get_by_latin_name("Homo sapiens")
    assert species is not None
    eq_(species.prefix, "HLA")


def test_species_get_by_latin_name_mouse():
    species = Species.get_by_latin_name("Mus musculus")
    assert species is not None
    eq_(species.prefix, "H2")


def test_species_get_by_latin_name_chicken():
    species = Species.get_by_latin_name("Gallus gallus")
    assert species is not None
    eq_(species.prefix, "Gaga")


def test_species_get_by_latin_name_unknown_returns_none():
    assert Species.get_by_latin_name("Nonexistus fictionalus") is None


def test_species_latin_name_property():
    species = Species.get_by_latin_name("Homo sapiens")
    eq_(species.latin_name, "Homo sapiens")
    eq_(species.name, "Homo sapiens")


def test_species_to_record_includes_latin_name():
    species = Species.get_by_latin_name("Homo sapiens")
    record = species.to_record()
    assert "species_latin_name" in record
    eq_(record["species_latin_name"], "Homo sapiens")
    # Backwards compatibility
    assert "species_name" in record
    eq_(record["species_name"], "Homo sapiens")


# ---------------------------------------------------------------------------
# 2. Unique alias regression tests
# ---------------------------------------------------------------------------


def test_species_get_unique_prefix_still_works():
    for prefix, expected_name in [
        ("HLA", "Homo sapiens"),
        ("H2", "Mus musculus"),
        ("BoLA", "Bos sp."),
    ]:
        species = Species.get(prefix)
        assert species is not None, f"Species.get({prefix!r}) returned None"
        eq_(species.name, expected_name)


def test_species_get_unique_common_name_still_works():
    for name in ["human", "mouse"]:
        species = Species.get(name)
        assert species is not None, f"Species.get({name!r}) returned None"


def test_parse_unique_prefix_strings_unchanged():
    """Representative existing cases still parse identically."""
    examples = [
        ("HLA-A*02:01", Allele),
        ("Gaga-BF1", Gene),
        ("Dare-UBA", Gene),
    ]
    for raw, expected_type in examples:
        result = parse(raw, raise_on_error=True)
        assert isinstance(result, expected_type), (
            f"{raw} parsed as {type(result)}, expected {expected_type}"
        )


def test_all_species_prefixes_and_other_prefixes_are_unique_after_normalization():
    owners_by_prefix = defaultdict(set)
    for latin_name, record in species_data.items():
        prefix = record.get("prefix")
        if prefix:
            owners_by_prefix[normalize_string(prefix)].add(latin_name)
        for alias in record.get("other prefixes", []) or []:
            owners_by_prefix[normalize_string(alias)].add(latin_name)

    collisions = {
        prefix: sorted(owners) for prefix, owners in owners_by_prefix.items() if len(owners) > 1
    }
    assert collisions == {}


# ---------------------------------------------------------------------------
# 3. Ambiguity tests using real runtime collision (Bubu)
# ---------------------------------------------------------------------------


def test_species_get_bubu_resolves_to_buffalo():
    """Bubu is now uniquely owned by water buffalo (eagle-owl is BuboBubo)."""
    species = Species.get("Bubu")
    assert species is not None
    eq_(species.name, "Bubalus bubalis")


def test_parse_bubu_dqa_resolves_water_buffalo():
    """Bubu-DQA should resolve to Bubalus bubalis."""
    result = parse("Bubu-DQA", raise_on_error=True)
    assert isinstance(result, Gene)
    eq_(result.species.name, "Bubalus bubalis")


def test_parse_bubobubo_dab1_resolves_eagle_owl():
    """BuboBubo-DAB1 should resolve to eagle-owl via long prefix."""
    result = parse("BuboBubo-DAB1", raise_on_error=True)
    assert isinstance(result, Gene)
    eq_(result.species.name, "Bubo bubo")


def test_parse_lanicola_species_resolves_fiscal_shrike():
    species = Species.get("LaniCola")
    assert species is not None
    eq_(species.name, "Lanius collaris")


def test_parse_bubu_allele_resolves_by_gene_context():
    """Bubu-DQA*01:01 should resolve to buffalo."""
    result = parse("Bubu-DQA*01:01", raise_on_error=True)
    assert isinstance(result, Allele)
    eq_(result.species.name, "Bubalus bubalis")


def test_species_get_by_latin_name_buffalo():
    """Direct latin name lookup bypasses ambiguity."""
    species = Species.get_by_latin_name("Bubalus bubalis")
    assert species is not None
    eq_(species.prefix, "Bubu")


def test_species_get_by_latin_name_eagle_owl():
    """Direct latin name lookup bypasses ambiguity."""
    species = Species.get_by_latin_name("Bubo bubo")
    assert species is not None
    eq_(species.prefix, "BuboBubo")


def test_long_prefix_always_parseable():
    """The 5+5 long prefix should always work for parsing any species."""
    for latin, long_prefix in [
        ("Bubo bubo", "BuboBubo"),
        ("Chrysemys picta", "ChrysPicta"),
        ("Gavialis gangeticus", "GaviaGange"),
        ("Casuarius casuarius", "CasuaCasua"),
        ("Caretta caretta", "CaretCaret"),
    ]:
        species = Species.get(long_prefix)
        assert species is not None, f"Species.get({long_prefix!r}) returned None"
        eq_(species.name, latin)


def test_generated_4_4_prefix_collisions_are_not_global_aliases():
    for alias in ["CaniLupu"]:
        assert Species.get(alias) is None
        assert Species.get_multiple(alias) == ()


def test_generated_5_5_prefix_collisions_fall_back_to_full_scientific_name():
    assert Species.get("CanisLupus") is not None
    eq_(Species.get("CanisLupus").name, "Canis lupus")
    assert Species.get("CanisLupusBaileyi") is None


def test_curated_prefix_keeps_global_owner_when_generated_alias_would_collide():
    for prefix, latin_name in [
        ("ChryPict", "Chrysemys picta"),
        ("LaniColl", "Lanius collurio"),
    ]:
        species = Species.get(prefix)
        assert species is not None
        eq_(species.name, latin_name)


def test_explicit_short_prefixes_beat_generated_collision_aliases():
    explicit_owners = defaultdict(set)
    for latin_name, record in raw_species_dict.items():
        explicit_owners[normalize_string(record["prefix"])].add(latin_name)
        for alias in record.get("other prefixes", []) or []:
            explicit_owners[normalize_string(alias)].add(latin_name)

    for generator in (_make_long_prefix, _make_5_5_prefix):
        for latin_name in raw_species_dict:
            alias = generator(latin_name)
            if not alias:
                continue
            owners = explicit_owners.get(normalize_string(alias))
            if not owners:
                continue
            species = Species.get(alias)
            assert species is not None
            assert species.name in owners


def test_trinomial_entries_do_not_get_full_concatenated_aliases():
    wolf = Species.get("CanisLupus")
    assert wolf is not None
    eq_(wolf.name, "Canis lupus")
    assert Species.get("CanisLupusBaileyi") is None


def test_long_prefix_works_for_allele_parsing():
    """HomoSapie-A*02:01 should parse as HLA-A*02:01."""
    result = parse("HomoSapie-A*02:01", raise_on_error=True)
    assert isinstance(result, Allele)
    eq_(result.species.prefix, "HLA")
    eq_(result.gene.name, "A")
    eq_(result.allele_fields, ("02", "01"))


def test_long_prefix_works_for_compact_allele():
    """HomoSapie-A0201 should also parse."""
    result = parse("HomoSapie-A0201", raise_on_error=True)
    assert isinstance(result, Allele)
    eq_(result.species.prefix, "HLA")
    eq_(result.allele_fields, ("02", "01"))


def test_full_latin_name_concatenated_works():
    """HomoSapiens-A*02:01 should parse (full latin name, no truncation)."""
    result = parse("HomoSapiens-A*02:01", raise_on_error=True)
    assert isinstance(result, Allele)
    eq_(result.species.prefix, "HLA")
    eq_(result.gene.name, "A")
    eq_(result.allele_fields, ("02", "01"))


def test_full_latin_name_concatenated_various_species():
    """Full concatenated latin names should work for any species."""
    for concat, expected_prefix in [
        ("DanioRerio-UBA", "Dare"),
        ("GallusGallus-BF1", "Gaga"),
        ("MusMusculus-K", "H2"),
    ]:
        result = parse(concat, raise_on_error=True)
        assert result is not None, f"parse({concat!r}) returned None"
        eq_(result.species.prefix, expected_prefix)


def test_latin_name_with_space_works():
    """'Homo sapiens-A*02:01' with a space in the species name should work."""
    result = parse("Homo sapiens-A*02:01", raise_on_error=True)
    assert isinstance(result, Allele)
    eq_(result.species.prefix, "HLA")


def test_decorated_scientific_name_falls_back_to_base_binomial():
    decorated = "Cyprinus carpio 'xingguonensis'"
    species = Species.get(decorated)
    assert species is not None
    eq_(species, Species.get("Cyprinus carpio"))


def test_parenthetical_scientific_name_falls_back_to_exact_subspecies_when_present():
    decorated = "Strix occidentalis caurina (northern spotted owl)"
    species = Species.get(decorated)
    assert species is not None
    eq_(species, Species.get("Strix occidentalis caurina"))


def test_unmodeled_trinomial_falls_back_to_base_binomial():
    species = Species.get("Canis lupus familiaris")
    assert species is not None
    eq_(species, Species.get("Canis lupus"))


def test_sourced_short_2_2_prefix_alias_resolves_species():
    for short_prefix, latin_name in [
        ("Abbr", "Abramis brama"),
        ("Crin", "Crocodylus intermedius"),
        ("Crjo", "Crocodylus johnstoni"),
        ("Crpa", "Crocodylus palustris"),
        ("Crrh", "Crocodylus rhombifer"),
        ("Euma", "Eublepharis macularius"),
        ("Geja", "Gekko japonicus"),
        ("Lili", "Limosa limosa"),
        ("Pato", "Parachondrostoma toxostoma"),
        ("Ruru", "Rutilus rutilus"),
        ("Spsp", "Spinus spinus"),
        ("Tycu", "Tympanuchus cupido"),
    ]:
        species = Species.get(short_prefix)
        assert species is not None, f"Species.get({short_prefix!r}) returned None"
        eq_(species.name, latin_name)


def test_sourced_short_2_2_prefix_alias_parses_genes():
    for raw, expected_prefix, expected_gene in [
        ("Abbr-DAB1", "AbraBram", "DAB1"),
        ("Crin-DB05", "CrocInte", "DB05"),
        ("Crjo-DB02", "CrocJohn", "DB02"),
        ("Crpa-DB02", "CrocPalu", "DB02"),
        ("Crrh-DB05", "CrocRhom", "DB05"),
        ("Lili-UA", "LimoLimo", "UA"),
        ("Pato-DAB1", "ParaToxo", "DAB1"),
        ("Ruru-DAB3", "RutiRuti", "DAB3"),
        ("Spsp-UA", "SpinSpin", "UA"),
    ]:
        result = parse(raw, raise_on_error=True)
        assert isinstance(result, Gene)
        eq_(result.species.prefix, expected_prefix)
        eq_(result.name, expected_gene)


def test_collision_backed_short_prefix_remains_blocked():
    # Hymo is source-backed for silver carp, but runtime already owns it.
    species = Species.get("Hymo")
    assert species is not None
    eq_(species.name, "Hylobates moloch")
    assert parse("Hymo-UA", raise_on_error=False) is None


def test_existing_prefix_owner_wins_over_source_backed_short_alias():
    # Orla is used by multiple fish datasets and also collides with orangutan.
    species = Species.get("Orla")
    assert species is not None
    eq_(species.name, "Pongo sp.")
    assert parse("Orla-UGA", raise_on_error=False) is None


def test_orla_still_parses_as_orangutan():
    result = parse("OrLA-A*01:01", raise_on_error=True)
    eq_(result.species.name, "Pongo sp.")
    eq_(result.gene.name, "A")


def test_ambiguous_or_unsourced_short_prefixes_are_not_added():
    for short_prefix in [
        "Moal",  # reused across Monopterus albus and Motacilla alba
    ]:
        assert Species.get(short_prefix) is None
        assert parse(f"{short_prefix}-DAB", raise_on_error=False) is None


def test_context_only_prefixes_stay_out_of_global_species_lookup():
    assert Species.get("Moal") is None
    assert Species.get("Motacilla alba") is not None
    eq_(Species.get("Motacilla alba").prefix, "MotaAlba")


def test_context_only_hymo_failure_is_informative():
    with pytest.raises(ParseError, match="Hypophthalmichthys molitrix"):
        parse("Hymo-DAB", raise_on_error=True)


def test_context_only_moal_failure_is_informative():
    with pytest.raises(ParseError, match="Monopterus albus"):
        parse("Moal-DAB", raise_on_error=True)
    with pytest.raises(ParseError, match="Motacilla alba"):
        parse("Moal-DAB", raise_on_error=True)


def test_context_only_orla_failure_is_informative():
    with pytest.raises(ParseError, match=r"Pongo sp\."):
        parse("ORLA-UAA", raise_on_error=True)
    with pytest.raises(ParseError, match="Oryzias latipes"):
        parse("ORLA-UAA", raise_on_error=True)


# ---------------------------------------------------------------------------
# 6. Failure modes for latin name / long prefix parsing
# ---------------------------------------------------------------------------


def test_nonexistent_latin_name_returns_none():
    assert parse("FictusSpecius-A*02:01", raise_on_error=False) is None


def test_partial_latin_name_too_short_returns_none():
    """Just 'Homo' should not resolve to a full allele."""
    assert parse("Homo-A*02:01", raise_on_error=False) is None


def test_misspelled_latin_name_returns_none():
    assert parse("HomoSapeins-A*02:01", raise_on_error=False) is None


def test_wrong_gene_for_species_returns_none():
    """HomoSapiens-BF1 — BF1 is a chicken gene, not human."""
    assert parse("HomoSapiens-BF1", raise_on_error=False) is None


# ---------------------------------------------------------------------------
# 4. Default species accepts latin name
# ---------------------------------------------------------------------------


def test_default_species_accepts_latin_name():
    """Parsing with default_species as latin name should work."""
    result = parse("A*02:01", default_species="Homo sapiens")
    assert result is not None
    assert isinstance(result, Allele)
    eq_(result.species.prefix, "HLA")


# ---------------------------------------------------------------------------
# 5. Prefixes that two species can legitimately derive
#
# The Klein 2+2 scheme (first two letters of genus + first two of species)
# produces collisions. Where IPD-MHC has designated a species, that species
# owns the prefix globally and the other keeps it as a context-only alias --
# the pattern already used for Hymo (Hylobates moloch vs the silver carp).
# https://github.com/pirl-unc/mhcgnomes/issues/112
# ---------------------------------------------------------------------------

# (prefix, species that resolves it globally, species holding it context-only)
#
# The global prefix goes to whichever side is actually used in print, not
# automatically to whichever side IPD-MHC designates. Caau-DAB and Caau-UFA
# appear in the cyprinid MHC literature while IPD holds no allele sequences
# for Canis aureus, so the goldfish resolves and the jackal's designation is
# reserved.
PREFIX_COLLISIONS = [
    ("Hymo", "Hylobates moloch", "Hypophthalmichthys molitrix"),
    ("Caau", "Carassius auratus", "Canis aureus"),
    ("Hyam", "Hybognathus amarus", "Hyperoodon ampullatus"),
]


@pytest.mark.parametrize("prefix,resolves_to,reserved_for", PREFIX_COLLISIONS)
def test_colliding_prefix_resolves_to_the_attested_species(prefix, resolves_to, reserved_for):
    species = Species.get(prefix)
    assert species is not None, prefix
    eq_(species.name, resolves_to)


@pytest.mark.parametrize("prefix,resolves_to,reserved_for", PREFIX_COLLISIONS)
def test_losing_side_keeps_prefix_as_context_only(prefix, resolves_to, reserved_for):
    from mhcgnomes.species import find_matching_context_only_species_objects

    names = {s.name for s in find_matching_context_only_species_objects(prefix)}
    assert reserved_for in names, (prefix, names)


@pytest.mark.parametrize("prefix,resolves_to,reserved_for", PREFIX_COLLISIONS)
def test_reserved_species_still_reachable_by_its_own_prefix(prefix, resolves_to, reserved_for):
    species = Species.get_by_latin_name(reserved_for)
    assert species is not None
    eq_(Species.get(species.prefix).name, reserved_for)


def test_published_goldfish_designations_still_parse():
    """
    Caau-DAB and Caau-UFA are published Carassius auratus designations, so
    Caau-* must keep resolving to the goldfish even though IPD-MHC designates
    Caau to Canis aureus.
    """
    eq_(Species.get("Caau").name, "Carassius auratus")
    for gene_name in ["DAB1", "DAB3", "UBA"]:
        eq_(parse(f"Caau-{gene_name}").species.name, "Carassius auratus")


def test_minnow_designations_still_parse():
    eq_(Species.get("Hyam").name, "Hybognathus amarus")
    eq_(parse("Hyam-DAB1").species.name, "Hybognathus amarus")


# ---------------------------------------------------------------------------
# 6. Historic prefixes must not be claimed twice
# https://github.com/pirl-unc/mhcgnomes/issues/110
# ---------------------------------------------------------------------------


def test_pren_resolves_to_the_langur():
    """
    Pren is Presbytis entellus, the former name of Semnopithecus entellus.
    Theropithecus gelada also carried it, which made the prefix ambiguous.
    """
    eq_(Species.get("Pren").name, "Semnopithecus entellus")
    eq_(parse("Pren").name, "Semnopithecus entellus")


def test_pren_class_one_agrees_with_bare_pren():
    eq_(parse("Pren class I").species, Species.get("Pren"))


def test_gelada_still_reachable_by_its_own_prefix():
    eq_(parse("Thge-DQA1").species.name, "Theropithecus gelada")


def test_no_prefix_is_claimed_by_two_species():
    """
    Canonical prefixes, historic 'old prefix' values and 'other prefixes' all
    feed the same global alias table, so none of them may name two species.
    Context-only prefixes are deliberately excluded -- they exist precisely to
    hold the losing side of a collision.
    """
    claims = defaultdict(set)
    for latin_name, record in species_data.items():
        for key in ("prefix", "old prefix"):
            value = record.get(key)
            if value:
                claims[normalize_string(value)].add(latin_name)
        for value in record.get("other prefixes", []) or []:
            claims[normalize_string(value)].add(latin_name)
    collisions = {p: sorted(v) for p, v in claims.items() if len(v) > 1}
    assert collisions == {}



#!/usr/bin/env python3
"""
deploy.py

Safe release-tag automation for this repository.

Policy (enforced):
  - Only deploy from REQUIRED_BRANCH (default: main)
  - Require a clean working tree (no staged/unstaged/untracked changes)
  - Require local REQUIRED_BRANCH to match REMOTE/REQUIRED_BRANCH
    (optionally after fetching, if --fetch is provided)
  - Use version already committed in VERSION_FILE (default: mhcgnomes/version.py)
  - Refuse to proceed if the tag already exists locally or on the remote

Does (in this order):
  - preflight checks (repo/branch/clean)
  - (optional) fetch remote branch
  - confirm local up-to-date with remote
  - python -m build
  - git tag v<version> (annotated) + push ONLY that tag

Publishing is handled by GitHub Actions after the tag push.

Dry-run:
  - Prints effectful commands but does NOT execute: build/tag/push/fetch.
  - Still runs read-only validations (e.g. tag existence checks, parsing version file).
"""

import argparse
import os
import shlex
import shutil
import subprocess
import sys
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import NoReturn, Optional

from release_utils import ReleaseError, parse_version_from_python_file, tag_for_version


def eprint(msg: str) -> None:
    print(msg, file=sys.stderr)


def die(msg: str, exit_code: int = 1) -> NoReturn:
    eprint(f"ERROR: {msg}")
    raise SystemExit(exit_code)


def note(msg: str) -> None:
    eprint(f"==> {msg}")


def ok(msg: str) -> None:
    eprint(f"OK: {msg}")


def shell_join(cmd: Sequence[str]) -> str:
    return " ".join(shlex.quote(c) for c in cmd)


@dataclass(frozen=True)
class Config:
    version_file: Path
    required_branch: str
    remote: str
    dry_run: bool
    fetch: bool
    env: Optional[dict[str, str]]


class CommandError(RuntimeError):
    def __init__(self, cmd: Sequence[str], returncode: int) -> None:
        super().__init__(f"Command failed (exit={returncode}): {shell_join(cmd)}")
        self.cmd = list(cmd)
        self.returncode = returncode


def require_nonempty(label: str, value: str) -> None:
    if not label:
        die("Internal error: label is empty in require_nonempty")
    if not value:
        die(f"{label} must not be empty")


def require_cmd(cmd: str, *, env: Optional[dict[str, str]] = None) -> None:
    require_nonempty("command", cmd)
    path = env.get("PATH") if env else None
    if shutil.which(cmd, path=path) is None:
        die(f"Missing required command: {cmd}")


def run_checked(
    cmd: Sequence[str],
    *,
    cwd: Optional[Path] = None,
    capture_stdout: bool = False,
    capture_stderr: bool = False,
    env: Optional[dict[str, str]] = None,
) -> subprocess.CompletedProcess:
    """
    Run a command (read-only checks, queries). Always executes.
    Always uses shell=False to avoid quoting/injection bugs.
    """
    note(shell_join(cmd))
    stdout = subprocess.PIPE if capture_stdout else None
    stderr = subprocess.PIPE if capture_stderr else None
    cp = subprocess.run(
        list(cmd),
        cwd=str(cwd) if cwd is not None else None,
        check=False,
        text=True,
        stdout=stdout,
        stderr=stderr,
        env=env,
    )
    if cp.returncode != 0:
        if cp.stdout:
            eprint(cp.stdout.rstrip("\n"))
        if cp.stderr:
            eprint(cp.stderr.rstrip("\n"))
        raise CommandError(cmd, cp.returncode)
    return cp


def run_effectful(
    cmd: Sequence[str], *, cwd: Path, dry_run: bool, env: Optional[dict[str, str]] = None
) -> None:
    """
    Run a command with side-effects (build/tag/push/fetch).
    In dry-run mode, only prints.
    """
    note(shell_join(cmd))
    if dry_run:
        return
    cp = subprocess.run(list(cmd), cwd=str(cwd), env=env)
    if cp.returncode != 0:
        raise CommandError(cmd, cp.returncode)


def git_output(cmd: Sequence[str], *, cwd: Path) -> str:
    cp = run_checked(cmd, cwd=cwd, capture_stdout=True, capture_stderr=True)
    assert cp.stdout is not None
    return cp.stdout.strip()


def git_succeeds(cmd: Sequence[str], *, cwd: Path) -> bool:
    cp = subprocess.run(
        list(cmd),
        cwd=str(cwd),
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    return cp.returncode == 0


def repo_root() -> Path:
    require_cmd("git")
    try:
        out = subprocess.check_output(["git", "rev-parse", "--show-toplevel"], text=True).strip()
    except subprocess.CalledProcessError:
        die("Not inside a git repository.")
    if not out:
        die("Failed to determine repository root.")
    root = Path(out)
    if not root.is_dir():
        die(f"Resolved repo root is not a directory: {root}")
    return root


def current_branch(*, cwd: Path) -> str:
    cp = subprocess.run(
        ["git", "symbolic-ref", "--quiet", "--short", "HEAD"],
        cwd=str(cwd),
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
        text=True,
    )
    if cp.returncode != 0:
        return ""
    return (cp.stdout or "").strip()


def ensure_on_branch(cfg: Config, *, cwd: Path) -> None:
    branch = current_branch(cwd=cwd)
    if not branch:
        die(
            "Detached HEAD (or unable to determine branch). "
            f"Check out '{cfg.required_branch}' and retry."
        )
    if branch != cfg.required_branch:
        die(f"Deploys must be run from '{cfg.required_branch}'. Current branch: '{branch}'.")
    ok(f"On branch '{cfg.required_branch}'")


def ensure_clean_tree(*, cwd: Path) -> None:
    status = git_output(["git", "status", "--porcelain=v1"], cwd=cwd)
    if status:
        eprint(status)
        die("Working tree not clean. Commit or stash changes before deploying.")
    ok("Working tree is clean")


def ensure_remote_exists(cfg: Config, *, cwd: Path) -> None:
    require_nonempty("remote", cfg.remote)
    if not git_succeeds(["git", "remote", "get-url", cfg.remote], cwd=cwd):
        die(f"Remote '{cfg.remote}' not found.")
    ok(f"Remote '{cfg.remote}' exists")


def validate_tag_name(tag: str, *, cwd: Path) -> None:
    require_nonempty("tag", tag)
    if not git_succeeds(
        ["git", "check-ref-format", "--allow-onelevel", f"refs/tags/{tag}"], cwd=cwd
    ):
        die(f"Tag '{tag}' is not a valid git tag name.")


def ensure_tag_absent(cfg: Config, tag: str, *, cwd: Path) -> None:
    validate_tag_name(tag, cwd=cwd)

    if git_succeeds(
        ["git", "show-ref", "--tags", "--quiet", "--verify", f"refs/tags/{tag}"], cwd=cwd
    ):
        die(f"Tag '{tag}' already exists locally.")

    if git_succeeds(
        ["git", "ls-remote", "--exit-code", "--tags", cfg.remote, f"refs/tags/{tag}"], cwd=cwd
    ):
        die(f"Tag '{tag}' already exists on {cfg.remote}.")

    ok(f"Tag '{tag}' does not exist (local or {cfg.remote})")


def maybe_fetch(cfg: Config, *, cwd: Path) -> None:
    if not cfg.fetch:
        note(
            f"Not fetching remotes (default). If refs are stale, rerun with --fetch or run: "
            f"git fetch {cfg.remote}"
        )
        return
    note(f"Fetching '{cfg.remote}/{cfg.required_branch}' (because --fetch was provided)...")
    run_effectful(
        ["git", "fetch", "--prune", cfg.remote, cfg.required_branch],
        cwd=cwd,
        dry_run=cfg.dry_run,
        env=cfg.env,
    )
    ok("Fetch complete")


def ensure_up_to_date(cfg: Config, *, cwd: Path) -> None:
    remote_refname = f"{cfg.remote}/{cfg.required_branch}"
    if not git_succeeds(["git", "rev-parse", "--verify", remote_refname], cwd=cwd):
        die(
            f"Cannot resolve {remote_refname}. Run: git fetch {cfg.remote} (or rerun with --fetch)."
        )

    local_ref = git_output(["git", "rev-parse", "--verify", "HEAD"], cwd=cwd)
    remote_ref = git_output(["git", "rev-parse", "--verify", remote_refname], cwd=cwd)
    base = git_output(["git", "merge-base", "HEAD", remote_refname], cwd=cwd)

    if local_ref != remote_ref:
        if base == local_ref:
            die(
                f"Local {cfg.required_branch} is BEHIND {remote_refname}. "
                f"Run: git pull --ff-only {cfg.remote} {cfg.required_branch}"
            )
        if base == remote_ref:
            die(
                f"Local {cfg.required_branch} is AHEAD of {remote_refname} (unpushed commits). "
                f"Push first: git push {cfg.remote} {cfg.required_branch}"
            )
        die(
            f"Local {cfg.required_branch} has DIVERGED from {remote_refname}. "
            f"Rebase/merge, push, then retry."
        )

    ok(f"Local {cfg.required_branch} matches {remote_refname}")


def clean_build_artifacts(*, cwd: Path, dry_run: bool) -> None:
    paths_to_remove = [
        cwd / "dist",
        cwd / "build",
        *sorted(cwd.glob("*.egg-info")),
    ]

    note("Cleaning build artifacts ...")
    for path in paths_to_remove:
        note(f"rm -rf {path}")
        if dry_run:
            continue
        shutil.rmtree(path, ignore_errors=True)


def build_distributions(cfg: Config, *, cwd: Path) -> None:
    require_cmd("python3", env=cfg.env)

    clean_build_artifacts(cwd=cwd, dry_run=cfg.dry_run)

    build_python = sys.executable or "python3"
    note("Checking build availability...")
    run_checked(
        [build_python, "-m", "build", "--version"],
        cwd=cwd,
        capture_stdout=True,
        capture_stderr=True,
    )

    note("Building distributions with python -m build (no isolation)...")
    run_effectful([build_python, "-m", "build", "--no-isolation"], cwd=cwd, dry_run=cfg.dry_run)
    ok("Build step complete")

    dist_dir = cwd / "dist"
    if not dist_dir.exists():
        die("dist/ directory does not exist after build.")
    dist_files = [p for p in sorted(dist_dir.iterdir()) if p.is_file()]
    if not dist_files:
        die("No files in dist/ after build.")
    ok("Distribution artifacts created")


def tag_and_push(cfg: Config, tag: str, version: str, *, cwd: Path) -> None:
    validate_tag_name(tag, cwd=cwd)
    require_nonempty("version", version)

    note(f"Creating annotated tag {tag}...")
    run_effectful(
        ["git", "tag", "-a", tag, "-m", f"Release {tag} ({version})"],
        cwd=cwd,
        dry_run=cfg.dry_run,
        env=cfg.env,
    )
    ok("Tag created")

    note(f"Pushing tag {tag} to {cfg.remote}...")
    run_effectful(
        ["git", "push", cfg.remote, f"refs/tags/{tag}"],
        cwd=cwd,
        dry_run=cfg.dry_run,
        env=cfg.env,
    )
    ok("Tag pushed")


def resolve_venv_bin(root: Path) -> Optional[Path]:
    for name in (".venv", "venv"):
        candidate = root / name / "bin"
        if (candidate / "python").exists() or (candidate / "python3").exists():
            return candidate
    return None


def build_run_env(venv_bin: Optional[Path]) -> dict[str, str]:
    env = os.environ.copy()
    if venv_bin is not None:
        env["PATH"] = f"{venv_bin}{os.pathsep}{env.get('PATH', '')}"
        env["VIRTUAL_ENV"] = str(venv_bin.parent)
    return env


def parse_args(argv: Sequence[str]) -> Config:
    p = argparse.ArgumentParser(prog="deploy.py")
    p.add_argument(
        "--dry-run", action="store_true", help="Print effectful commands; do not execute them."
    )
    p.add_argument(
        "--fetch", action="store_true", help="Fetch remote branch before comparing refs."
    )
    p.add_argument("--version-file", default="mhcgnomes/version.py")
    p.add_argument("--required-branch", default="main")
    p.add_argument("--remote", default="origin")

    ns = p.parse_args(list(argv))
    require_nonempty("--required-branch", ns.required_branch)
    require_nonempty("--remote", ns.remote)

    return Config(
        version_file=Path(ns.version_file),
        required_branch=ns.required_branch,
        remote=ns.remote,
        dry_run=bool(ns.dry_run),
        fetch=bool(ns.fetch),
        env=None,
    )


def main(argv: Sequence[str]) -> int:
    cfg = parse_args(argv)
    require_cmd("git")

    root = repo_root()
    venv_bin = resolve_venv_bin(root)
    run_env = build_run_env(venv_bin)

    # Interpret configured paths relative to repo root so deploy.sh works from anywhere.
    cfg = Config(
        version_file=(root / cfg.version_file).resolve(),
        required_branch=cfg.required_branch,
        remote=cfg.remote,
        dry_run=cfg.dry_run,
        fetch=cfg.fetch,
        env=run_env,
    )

    ok(f"Repo root: {root}")
    if venv_bin is not None:
        ok(f"Using venv: {venv_bin.parent}")
    else:
        note("No venv found; using system PATH")

    # Preflight: cheap local checks first.
    ensure_remote_exists(cfg, cwd=root)
    ensure_on_branch(cfg, cwd=root)
    ensure_clean_tree(cwd=root)

    # Now ensure branch is synced with remote (optionally fetching).
    maybe_fetch(cfg, cwd=root)
    ensure_up_to_date(cfg, cwd=root)

    try:
        version = parse_version_from_python_file(cfg.version_file)
    except ReleaseError as exc:
        die(str(exc))
    tag = tag_for_version(version)
    validate_tag_name(tag, cwd=root)

    note("Preparing release tag:")
    note(f"  version file: {cfg.version_file}")
    note(f"  version:      {version}")
    note(f"  tag:          {tag}")
    note(f"  branch:       {cfg.required_branch}")
    note(f"  remote:       {cfg.remote}")
    if cfg.dry_run:
        note("  dry-run:      enabled")

    ensure_tag_absent(cfg, tag, cwd=root)

    if cfg.dry_run:
        note("Dry run summary:")
        note("  would run: python -m build --no-isolation")
        note(f"  would run: git tag -a {tag} -m 'Release {tag} ({version})'")
        note(f"  would run: git push {cfg.remote} refs/tags/{tag}")
        ok("Dry run complete")
        return 0

    build_distributions(cfg, cwd=root)
    tag_and_push(cfg, tag, version, cwd=root)
    ok(f"Release tag pushed: {tag}")
    note("GitHub Actions will build and publish this tag to PyPI.")
    return 0


if __name__ == "__main__":
    try:
        raise SystemExit(main(sys.argv[1:]))
    except CommandError as e:
        die(str(e), exit_code=e.returncode)



from mhcgnomes import Gene, Species, parse

from .common import eq_


def test_species_lineage_helpers():
    bos = Species.get("Bos sp.")
    bota = Species.get("Bos taurus")
    galliformes = Species.get("Galliformes sp.")
    gaga = Species.get("Gallus gallus")

    assert bos is not None
    assert bota is not None
    assert galliformes is not None
    assert gaga is not None

    assert bos.is_parent_of(bota)
    assert bota.is_child_of(bos)
    assert bos.is_ancestor_of(bota)
    assert bota.is_descendant_of(bos)
    assert galliformes.is_ancestor_of(gaga)
    assert gaga.is_descendant_of(galliformes)


def test_species_strict_exact_match():
    result = parse("HLA-A*02:01", species="Homo sapiens")
    eq_(result.species.name, "Homo sapiens")


def test_species_strict_accepts_decorated_subspecies_name():
    result = parse(
        "Stoc-DAB1",
        species="Strix occidentalis caurina (northern spotted owl)",
        raise_on_error=True,
    )
    eq_(result.species.name, "Strix occidentalis caurina")


def test_species_strict_bola_converts_to_bos_taurus():
    result = parse("BoLA-DRB3*01:01", species="Bos taurus")
    eq_(result.species.name, "Bos taurus")
    eq_(result.to_string(), "Bota-DRB3*01:01")


def test_species_strict_sla_converts_to_sus_scrofa():
    result = parse("SLA-1*01:01", species="Sus scrofa")
    eq_(result.species.name, "Sus scrofa")
    eq_(result.to_string(), "Susc-1*01:01")


def test_species_strict_dla_converts_to_canis_lupus():
    result = parse("DLA-DRB1", species="Canis lupus")
    eq_(result.species.name, "Canis lupus")
    eq_(result.to_string(), "Calu-DRB1")


def test_species_strict_bola_with_bos_sp():
    result = parse("BoLA-DRB3*01:01", species="Bos sp.")
    eq_(result.species.name, "Bos sp.")
    eq_(result.to_string(), "BoLA-DRB3*01:01")


def test_species_strict_hla_with_bos_taurus_errors():
    assert parse("HLA-A*02:01", species="Bos taurus", raise_on_error=False) is None


def test_species_strict_bola_with_human_errors():
    assert parse("BoLA-DRB3*01:01", species="Homo sapiens", raise_on_error=False) is None


def test_species_strict_bola_with_sheep_errors():
    assert parse("BoLA-DRB3*01:01", species="Ovis sp.", raise_on_error=False) is None


def test_species_strict_taxonomic_node_does_not_convert_to_descendant():
    assert parse("Galliformes", species="Gallus gallus", raise_on_error=False) is None


def test_species_strict_child_result_not_cast_up_to_parent():
    assert parse("Gaga-BF1", species="Galliformes sp.", raise_on_error=False) is None


def test_species_strict_context_only_hymo_rescues_silver_carp():
    result = parse("Hymo-DAB", species="Hypophthalmichthys molitrix", raise_on_error=True)
    assert isinstance(result, Gene)
    eq_(result.species.name, "Hypophthalmichthys molitrix")
    eq_(result.to_string(), "HypoMoli-DAB")


def test_species_strict_context_only_orla_rescues_medaka():
    result = parse("ORLA-UAA", species="Oryzias latipes", raise_on_error=True)
    assert isinstance(result, Gene)
    eq_(result.species.name, "Oryzias latipes")
    eq_(result.to_string(), "OryzLati-UAA")


def test_species_strict_context_only_orla_rescues_killifish():
    result = parse("Orla-UAA", species="Iconisemion striatum", raise_on_error=True)
    assert isinstance(result, Gene)
    eq_(result.species.name, "Iconisemion striatum")
    eq_(result.to_string(), "IconStri-UAA")


def test_species_strict_context_only_moal_rescues_swamp_eel():
    result = parse("Moal-DAB", species="Monopterus albus", raise_on_error=True)
    assert isinstance(result, Gene)
    eq_(result.species.name, "Monopterus albus")
    eq_(result.to_string(), "MonoAlbu-DAB")


def test_species_strict_context_only_moal_rescues_white_wagtail():
    result = parse("Moal-DAB4", species="Motacilla alba", raise_on_error=True)
    assert isinstance(result, Gene)
    eq_(result.species.name, "Motacilla alba")
    eq_(result.to_string(), "MotaAlba-DAB4")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Sequence
from dataclasses import dataclass
from typing import Union

from .result_with_multiple_alleles import ResultWithMultipleAlleles
from .result_with_species import ResultWithSpecies
from .species import Species


@dataclass(eq=False, repr=False, frozen=True, init=False)
class AmbiguousAlleles(ResultWithMultipleAlleles):
    """
    A set of distinct allele designations that an input string cannot
    disambiguate between. Used for IPD-MHC style slash designations
    such as "SahaI*74/88" (Caldwell et al. 2018, PMC6092122), where an
    observed sequence matches two database entries because they are
    identical in the region that was typed.

    Unlike Serotype or Haplotype, the grouping here does not reflect a
    biological or serological relationship — just typing ambiguity.
    """

    def __init__(
        self,
        species: Species,
        alleles: Sequence[ResultWithSpecies],
        raw_string: Union[str, None] = None,
    ):
        name = "/".join(_short_name(a) for a in alleles)
        ResultWithMultipleAlleles.__init__(
            self, species=species, name=name, alleles=alleles, raw_string=raw_string
        )

    def to_string(self, include_species=True, use_old_species_prefix=False):
        parts = [
            a.to_string(include_species=False, use_old_species_prefix=use_old_species_prefix)
            for a in self.alleles
        ]
        joined = "/".join(parts)
        if include_species:
            species_str = self.species.to_string(
                include_species=include_species,
                use_old_species_prefix=use_old_species_prefix,
            )
            return f"{species_str}-{joined}"
        return joined

    def compact_string(self, include_species=False, use_old_species_prefix=False):
        return self.to_string(
            include_species=include_species, use_old_species_prefix=use_old_species_prefix
        )

    def to_record(self):
        d = self.species.to_record()
        d["ambiguous_alleles"] = self.to_string()
        return d

    @property
    def mhc_class(self):
        # Share a single mhc_class across members when they agree (e.g.
        # all class I); otherwise None.
        classes = {getattr(a, "mhc_class", None) for a in self.alleles}
        classes.discard(None)
        if len(classes) == 1:
            return next(iter(classes))
        return None

    @property
    def has_mhc_class(self):
        return self.mhc_class is not None


def _short_name(allele):
    if hasattr(allele, "compact_string"):
        return allele.compact_string(include_species=False)
    return getattr(allele, "name", str(allele))



"""
Tests for species, gene aliases, and gene series added in v3.22.0.

Covers:
- Ctenomys tuco-tuco rodents (4 species with numbered DRB lineages)
- New birds: herons, egrets, petrel, penguin
- New fish: minnow, bream, pike
- New lizard: Ameiva ameiva
- Crocodilian DB series extension (DB10-DB55)
- Turdus thrush PFA series (PFA01-PFA30)
"""

import pytest

from mhcgnomes import Allele, Gene, Species, parse

from .common import eq_

# ---------------------------------------------------------------------------
# Ctenomys tuco-tuco rodents
# ---------------------------------------------------------------------------

CTENOMYS_SPECIES = [
    ("Ctta", "Ctenomys talarum"),
    ("Ctau", "Ctenomys australis"),
    ("Ctto", "Ctenomys torquatus"),
    ("Ctpe", "Ctenomys pearsoni"),
]


@pytest.mark.parametrize("prefix, latin", CTENOMYS_SPECIES)
def test_ctenomys_species_parseable(prefix, latin):
    species = Species.get(prefix)
    assert species is not None, f"Species not found for prefix {prefix}"
    assert species.latin_name == latin


@pytest.mark.parametrize("prefix, latin", CTENOMYS_SPECIES)
def test_ctenomys_latin_name_lookup(prefix, latin):
    species = Species.get(latin)
    assert species is not None, f"Species not found for latin name {latin}"


def test_ctenomys_talarum_DRB01():
    # DRB01 is an allele lineage (DRB*01), not a separate gene
    eq_(parse("Ctta-DRB01"), Allele.get("Ctta", "DRB", "01"))


def test_ctenomys_talarum_DQB():
    eq_(parse("Ctta-DQB"), Gene.get("Ctta", "DQB"))


def test_ctenomys_pearsoni_DRB03():
    # DRB03 is an allele lineage (DRB*03), not a separate gene
    eq_(parse("Ctpe-DRB03"), Allele.get("Ctpe", "DRB", "03"))


def test_ctenomys_torquatus_inherits_rodentia_DRB1():
    """Ctenomys species should also inherit standard DRB1 from Rodentia sp."""
    gene = Gene.get("Ctta", "DRB1")
    assert gene is not None


# ---------------------------------------------------------------------------
# New bird species
# ---------------------------------------------------------------------------

NEW_BIRDS = [
    ("ArdeCine", "Ardea cinerea"),
    ("HaloCaer", "Halobaena caerulea"),
    ("BubuIbis", "Bubulcus ibis"),
    ("EgreSacr", "Egretta sacra"),
    ("EudyChrys", "Eudyptes chrysocome"),
]


@pytest.mark.parametrize("prefix, latin", NEW_BIRDS)
def test_new_bird_species_parseable(prefix, latin):
    species = Species.get(prefix)
    assert species is not None, f"Species not found for prefix {prefix}"
    assert species.latin_name == latin


@pytest.mark.parametrize("prefix, latin", NEW_BIRDS)
def test_new_bird_latin_name_lookup(prefix, latin):
    species = Species.get(latin)
    assert species is not None


def test_new_bird_short_prefix_lookup():
    """Short prefixes in 'other prefixes' should also resolve."""
    assert Species.get("Arci") is not None
    assert Species.get("Haca") is not None
    assert Species.get("Buib") is not None
    assert Species.get("Egsa") is not None
    assert Species.get("Euch") is not None


def test_ardea_cinerea_inherits_DAB1():
    eq_(parse("Arci-DAB1"), Gene.get("Arci", "DAB1"))


def test_egretta_sacra_inherits_DAB():
    eq_(parse("Egsa-DAB"), Gene.get("Egsa", "DAB"))


def test_eudyptes_chrysocome_HA1F():
    eq_(parse("Euch-HA1F"), Gene.get("Euch", "HA1F"))


def test_eudyptes_chrysocome_DMB():
    eq_(parse("Euch-DMB"), Gene.get("Euch", "DMB"))


# ---------------------------------------------------------------------------
# New fish species
# ---------------------------------------------------------------------------

NEW_FISH = [
    ("HyboAmar", "Hybognathus amarus"),
    ("MegaAmbl", "Megalobrama amblycephala"),
    ("EsoxLuci", "Esox lucius"),
]


@pytest.mark.parametrize("prefix, latin", NEW_FISH)
def test_new_fish_species_parseable(prefix, latin):
    species = Species.get(prefix)
    assert species is not None, f"Species not found for prefix {prefix}"
    assert species.latin_name == latin


def test_hybognathus_amarus_inherits_DAB1():
    eq_(parse("Hyam-DAB1"), Gene.get("Hyam", "DAB1"))


def test_megalobrama_inherits_UA():
    eq_(parse("Meam-UA"), Gene.get("Meam", "UA"))


def test_esox_lucius_HA2B():
    eq_(parse("Eslu-HA2B"), Gene.get("Eslu", "HA2B"))


def test_esox_lucius_HB2D():
    eq_(parse("Eslu-HB2D"), Gene.get("Eslu", "HB2D"))


# ---------------------------------------------------------------------------
# New reptile: Ameiva ameiva
# ---------------------------------------------------------------------------


def test_ameiva_species_parseable():
    species = Species.get("AmeiAmei")
    assert species is not None
    assert species.latin_name == "Ameiva ameiva"


def test_ameiva_short_prefix():
    assert Species.get("Amam") is not None


def test_ameiva_LC1():
    eq_(parse("Amam-LC1"), Gene.get("Amam", "LC1"))


def test_ameiva_LC13():
    eq_(parse("Amam-LC13"), Gene.get("Amam", "LC13"))


# ---------------------------------------------------------------------------
# Crocodilian DB series extension (DB10-DB55)
# ---------------------------------------------------------------------------


def test_crocodylia_DB10_defined():
    """DB10 should be inherited by all crocodilians."""
    gene = Gene.get("Crpo", "DB10")
    assert gene is not None


def test_crpo_DB50():
    eq_(parse("Crpo-DB50"), Gene.get("Crpo", "DB50"))


def test_crpo_DB55():
    eq_(parse("Crpo-DB55"), Gene.get("Crpo", "DB55"))


def test_crni_inherits_DB30():
    """Nile crocodile should also inherit extended DB series."""
    gene = Gene.get("CrocNilo", "DB30")
    assert gene is not None


# ---------------------------------------------------------------------------
# Turdus thrush PFA series (PFA01-PFA30)
# ---------------------------------------------------------------------------


def test_turdus_sp_defined():
    species = Species.get("Turdus")
    assert species is not None


def test_turdus_merula_parent_is_turdus_sp():
    species = Species.get("TurdMeru")
    assert species is not None
    assert species.parent.latin_name == "Turdus sp."


def test_turdus_merula_PFA01():
    eq_(parse("TurdMeru-PFA01"), Gene.get("TurdMeru", "PFA01"))


def test_turdus_merula_PFA15():
    eq_(parse("TurdMeru-PFA15"), Gene.get("TurdMeru", "PFA15"))


def test_turdus_merula_PFA30():
    eq_(parse("TurdMeru-PFA30"), Gene.get("TurdMeru", "PFA30"))


def test_turdus_naumanni_inherits_PFA():
    gene = Gene.get("TurdNaum", "PFA10")
    assert gene is not None


def test_turdus_merula_still_inherits_aves_DAB():
    """Re-parenting to Turdus sp. should not break Aves inheritance."""
    gene = Gene.get("TurdMeru", "DAB")
    assert gene is not None
